import unicodedata
from dataclasses import dataclass
from typing import Dict, List, Tuple

from whylogs.core.metrics.compound_metric import CompoundMetric
from whylogs.core.metrics.metrics import (
    DistributionMetric,
    MetricConfig,
    OperationResult,
)
from whylogs.core.preprocessing import PreprocessedColumn
from whylogs.core.proto import MetricMessage

_STRING_LENGTH = "string_length"


@dataclass
class UnicodeRangeMetric(CompoundMetric):
    """
    For string values, maintains a DistributionMetric for the counts of
    characters that fall within user-defined codepoint ranges.

    Parameters
     ----------
     range_definitions : Dict[str, Tuple[int, int]]
         Defines the character ranges to be counted. The key servers as
         the range name and should only contain alphanumeric, hyphen, and
         underscore characters. The tuple defines the Unicode codepoint
         range to be tracked. The string length is tracked under the key
         "STRING_LENGTH" so don't use that as a range name.
    """

    range_definitions: Dict[str, Tuple[int, int]]
    lower_case: bool = True
    normalize: bool = True

    def __post_init__(self):
        super(type(self), self).__post_init__()
        self.range_definitions["UNKNOWN"] = (0, 0)  # catchall for characters not in a defined range
        for key, range in self.range_definitions.items():
            if range[0] > range[1]:
                raise ValueError(f"Invalid codepoint range {key}")
            if range[0] < 0 or 0x10FFFF < range[1]:
                raise ValueError(f"Invalid codepoint range {key}")
            if ":" in key or "/" in key:
                raise ValueError(f"Invalid range name {key}")
        if _STRING_LENGTH in self.range_definitions:
            raise ValueError("STRING_LENGTH cannot be used as a range name")

        submetrics = {
            key: DistributionMetric.zero(MetricConfig(large_kll_k=False)) for key in self.range_definitions.keys()
        }
        submetrics[_STRING_LENGTH] = DistributionMetric.zero(MetricConfig(large_kll_k=False))
        super(type(self), self).__init__(submetrics)  # type: ignore

    @property
    def namespace(self) -> str:
        return "unicode_range"

    def merge(self, other: "UnicodeRangeMetric") -> "UnicodeRangeMetric":
        submetrics = self.merge_submetrics(other)
        result = UnicodeRangeMetric(self.range_definitions)
        result.submetrics = submetrics
        return result

    def columnar_update(self, view: PreprocessedColumn) -> OperationResult:
        data = (
            view.pandas.strings.to_list() if view.pandas.strings is not None and not view.pandas.strings.empty else []
        )
        data = (data + view.list.strings) if view.list.strings else data
        range_data: Dict[str, List[int]] = {range_name: [] for range_name in self.range_definitions.keys()}
        lengths: List[int] = []
        for value in data:
            lengths.append(len(value))
            range_counter: Dict[str, int] = {range_name: 0 for range_name in self.range_definitions.keys()}
            # TODO: need to transform to utf-32 or handle surrogates
            s = unicodedata.normalize("NFD", value) if self.normalize else value
            s = s.lower() if self.lower_case else s
            for char in s:
                found = False
                for range_name, range_limits in self.range_definitions.items():
                    if range_limits[0] <= ord(char) <= range_limits[1]:
                        range_counter[range_name] += 1
                        found = True
                if not found:
                    range_counter["UNKNOWN"] += 1

            for range_name, range_count in range_counter.items():
                range_data[range_name].append(range_count)

        submetric_col = PreprocessedColumn()
        submetric_col.list.ints = lengths
        self.submetrics[_STRING_LENGTH].columnar_update(submetric_col)
        for range_name, range_list in range_data.items():
            submetric_col.list.ints = range_list
            self.submetrics[range_name].columnar_update(submetric_col)
        return OperationResult.ok(len(data))

    @classmethod
    def zero(cls, config: MetricConfig) -> "UnicodeRangeMetric":
        return cls(config.unicode_ranges, lower_case=config.lower_case, normalize=config.normalize)

    @classmethod
    def from_protobuf(cls, msg: MetricMessage) -> "UnicodeRangeMetric":
        submetrics = cls.submetrics_from_protobuf(msg)
        # The MetricMessage doesn't contain the range definitions, so we preserve
        # the range names with empty bounds
        ranges = {sub_name: (0, 0) for sub_name in submetrics.keys()}
        del ranges[_STRING_LENGTH]
        result = cls(ranges)
        result.submetrics = submetrics
        return result



import os
from dataclasses import dataclass
from datetime import date, datetime, timedelta, timezone
from importlib import resources
from logging import getLogger
from typing import Iterable, Optional, Tuple, Union

import pandas as pd

from whylogs.datasets.base import Batch, Dataset
from whylogs.datasets.configs import BaseConfig, DatasetConfig, EcommerceConfig
from whylogs.datasets.utils import (
    _adjust_df_date,
    _get_dataset_path,
    _parse_interval,
    _validate_timestamp,
)

logger = getLogger(__name__)


base_config = BaseConfig()


@dataclass(init=False)
class Ecommerce(Dataset):
    """Ecommerce Dataset"""

    baseline_df: pd.DataFrame
    inference_df: pd.DataFrame
    inference_interval: str = "1d"
    number_days: int = 1
    unit: str = "D"
    url: str = EcommerceConfig.url
    baseline_timestamp: Union[date, datetime] = datetime.now(timezone.utc).replace(
        hour=0, minute=0, second=0, microsecond=0
    )
    inference_start_timestamp: Union[date, datetime] = datetime.now(timezone.utc).replace(
        hour=0, minute=0, second=0, microsecond=0
    ) + timedelta(days=1)
    original: bool = False
    dataset_config: DatasetConfig = EcommerceConfig()

    def __init__(self, version: str = "base") -> None:
        """Initializes internal dataframes.

        If the files are already present locally, won't try to download from S3.

        Parameters
        ----------
        version : str, optional
            The desired dataset's version, by default "in_domain"

        """
        if version not in self.dataset_config.available_versions:
            raise ValueError("Version not found in list of available versions.")
        self.version = version

        baseline_file = os.path.join(
            _get_dataset_path(self.dataset_config.folder_name), "baseline_dataset_{}.csv".format(self.version)
        )
        inference_file = os.path.join(
            _get_dataset_path(self.dataset_config.folder_name), "inference_dataset_{}.csv".format(self.version)
        )

        try:
            self.baseline_df = pd.read_csv(baseline_file)
            self.inference_df = pd.read_csv(inference_file)
        except FileNotFoundError:
            self.baseline_df = pd.read_csv("{}/baseline_dataset_{}.csv".format(self.url, self.version))
            self.baseline_df.to_csv(baseline_file, index=False)

            self.inference_df = pd.read_csv("{}/inference_dataset_{}.csv".format(self.url, self.version))
            self.inference_df.to_csv(inference_file, index=False)

        self.baseline_df = _adjust_df_date(self.baseline_df, new_start_date=self.baseline_timestamp)
        self.inference_df = _adjust_df_date(self.inference_df, new_start_date=self.inference_start_timestamp)

    @classmethod
    def describe_versions(cls) -> Tuple[str]:
        available_versions = cls.dataset_config.available_versions
        return available_versions

    @classmethod
    def describe(cls) -> Optional[str]:
        descr = resources.read_text(base_config.description_folder, cls.dataset_config.description_file)
        return descr

    def get_baseline(self) -> Batch:
        data = self.baseline_df
        baseline = Batch(
            timestamp=self.baseline_timestamp, data=data, dataset_config=self.dataset_config, version=self.version
        )
        return baseline

    def _truncate_and_check_timezone(self, timestamp: datetime) -> datetime:
        if timestamp.tzinfo is None:
            logger.warning("No timezone set in the datetime_timestamp object. Default to local timezone")
            timestamp = timestamp.astimezone(tz=timezone.utc)
        timestamp = timestamp.replace(hour=0, minute=0, second=0, microsecond=0)
        return timestamp

    def _validate_interval(self, interval: str) -> Tuple[int, str]:
        """Checks if desired interval are of acceptable units and inside maximum duration limits."""
        number_days, unit = _parse_interval(interval)
        if number_days > self.dataset_config.max_interval:
            raise ValueError("Maximum allowed interval for this dataset is {}".format(self.dataset_config.max_interval))
        if unit != "D":
            raise ValueError("Current accepted unit for this dataset is {}".format(self.dataset_config.base_unit))
        return (number_days, unit)

    def get_inference_data(
        self, target_date: Optional[Union[date, datetime]] = None, number_batches: Optional[int] = None
    ) -> Union[Batch, Iterable[Batch]]:
        """Get batch(es) from inference dataset.

        Parameters
        ----------
        target_date : Optional[Union[date, datetime]], optional
            Target date for single batch. If datetime is passed, only date will be considered, by default None
        number_batches : Optional[int], optional
            Number of batches to be retrieved. Each batch will have a time interval as defined by `inference_interval` from `set_parameters`. By default None

        Returns
        -------
        Union[Batch, Iterable[Batch]]
            Can return a single batch or an interator of batches, depending on input parameters
        """
        if not target_date and not number_batches:
            raise ValueError("date or number_batches must be passed to get_inference_data.")
        if target_date and number_batches:
            raise ValueError("Either date or number_batches should be passed, not both.")
        if target_date and isinstance(target_date, (date, datetime)):
            _date: datetime = _validate_timestamp(target_date)
            _date = self._truncate_and_check_timezone(_date)
            mask = self.inference_df["date"] == _date
            data = self.inference_df.loc[mask]
            inference = Batch(timestamp=_date, data=data, dataset_config=self.dataset_config, version=self.version)
            return inference
        if number_batches:
            batches = EcommerceDatasetIterator(
                self.inference_df,
                number_days=self.number_days,
                number_batches=number_batches,
                version=self.version,
                config=self.dataset_config,
            )
            return batches
        raise ValueError("Target date should be either of date or datetime type.")

    def set_parameters(
        self,
        inference_interval: Optional[str] = None,
        baseline_timestamp: Optional[Union[date, datetime]] = None,
        inference_start_timestamp: Optional[Union[date, datetime]] = None,
        original: Optional[bool] = None,
    ) -> None:
        """Set timestamp and interval parameters for the dataset object.

        Parameters
        ----------
        inference_interval : Optional[str], optional
            Interval for the inference batches. If none is passed, daily inference batches will be returned, by default None
        baseline_timestamp : Optional[Union[date, datetime]], optional
            Timestamp for the baseline dataset. If none is passed, timestamp will be equal to the current day, by default None
        inference_start_timestamp : Optional[Union[date, datetime]], optional
            Timestamp for the start of the inference dataset. If none is passed, timestamp will be equal to tomorrow's date, by default None
        original : Optional[bool], optional
            _If true, sets both baseline and inference timestamps to the dataset's original timestamp, by default None
        """
        if inference_interval:
            self.inference_interval = inference_interval
            self.number_days, self.unit = self._validate_interval(self.inference_interval)
        if original:
            self.baseline_timestamp = self.dataset_config.baseline_start_timestamp[self.version]
            self.inference_start_timestamp = self.dataset_config.inference_start_timestamp[self.version]
            self.inference_df = _adjust_df_date(self.inference_df, self.inference_start_timestamp)
            self.baseline_df = _adjust_df_date(self.baseline_df, new_start_date=self.baseline_timestamp)

        if baseline_timestamp:
            if not original:
                _baseline_date: datetime = _validate_timestamp(baseline_timestamp)
                _baseline_date = self._truncate_and_check_timezone(_baseline_date)
                self.baseline_timestamp = _baseline_date
                self.baseline_df = _adjust_df_date(self.baseline_df, self.baseline_timestamp)

            else:
                logger.warning(
                    "baseline_timestamp and inference_start_timestamp overriden by original timestamps due to original = True"
                )

        if inference_start_timestamp:
            if not original:
                _inference_date: datetime = _validate_timestamp(inference_start_timestamp)
                _inference_date = self._truncate_and_check_timezone(_inference_date)
                self.inference_start_timestamp = _inference_date
                self.inference_df = _adjust_df_date(self.inference_df, self.inference_start_timestamp)
            else:
                logger.warning(
                    "baseline_timestamp and inference_start_timestamp overriden by original timestamps due to original = True"
                )


class EcommerceDatasetIterator:
    """Iterator to retrieve inference batches, when multiple batches are required."""

    def __init__(
        self, df: pd.DataFrame, number_days: int, number_batches: int, version: str, config=DatasetConfig
    ) -> None:
        self._df = df
        self._number_days = number_days
        self._number_batches = number_batches
        self.version = version
        self.config = config

    def __iter__(self):
        self._index: date = self._df.iloc[0].name
        self._batch_counter: int = 0
        return self

    def __next__(self) -> Batch:
        if self._batch_counter >= self._number_batches:
            raise StopIteration
        if self._index <= self._df.iloc[-1].name:
            day = self._index
            data = self._df[day : day + timedelta(days=(self._number_days - 1))]  # type: ignore
            inference = Batch(timestamp=day, data=data, dataset_config=self.config, version=self.version)
            self._index = data.index[-1] + timedelta(days=1)
            self._batch_counter += 1

            return inference

        raise StopIteration



from .cardinality_metrics import distinct_number_in_range
from .count_metrics import (
    count_below_number,
    null_percentage_below_number,
    null_values_below_number,
)
from .distribution_metrics import (
    greater_than_number,
    mean_between_range,
    quantile_between_range,
    smaller_than_number,
    stddev_between_range,
)
from .frequent_items import (
    frequent_strings_in_reference_set,
    n_most_common_items_in_set,
)

ALL = [
    greater_than_number,
    smaller_than_number,
    mean_between_range,
    stddev_between_range,
    quantile_between_range,
    count_below_number,
    null_values_below_number,
    null_percentage_below_number,
    distinct_number_in_range,
    frequent_strings_in_reference_set,
    n_most_common_items_in_set,
]



from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List

from whylogs.core.metrics.condition_count_metric import Condition
from whylogs.core.preprocessing import PreprocessedColumn


class Validator(ABC):
    name: str
    conditions: Dict[str, Condition]
    actions: List[Callable[[str, str, Any], None]]

    @abstractmethod
    def columnar_validate(self, data: PreprocessedColumn):
        pass

    @abstractmethod
    def to_summary_dict(self) -> Dict[str, Any]:
        pass



from typing import Optional, Tuple, Union

from whylogs.core.metrics import DistributionMetric
from whylogs.core.view.column_profile_view import ColumnProfileView


def get_distribution_metrics(
    column_view: Optional[ColumnProfileView],
) -> Union[Tuple[None, None, None], Tuple[float, float, float]]:
    if column_view is None:
        raise ValueError("Column view not found.")
    distribution_metric: Optional[DistributionMetric] = column_view.get_metric("distribution")
    if distribution_metric is None:
        return None, None, None

    min_val = distribution_metric.kll.value.get_min_value()
    max_val = distribution_metric.kll.value.get_max_value()
    range_val = max_val - min_val
    return min_val, max_val, range_val



import itertools
import logging
from dataclasses import dataclass
from typing import Any, Iterable, Iterator, List, Optional

from whylogs.core.stubs import np as np
from whylogs.core.stubs import pd as pd

logger = logging.getLogger("whylogs.core.views")

try:
    import pandas.core.dtypes.common as pdc
except:  # noqa
    pass


@dataclass
class ListView:
    ints: Optional[List[int]] = None
    floats: Optional[List[float]] = None
    strings: Optional[List[str]] = None
    objs: Optional[List[Any]] = None

    def iterables(self) -> List[List[Any]]:
        it_list = []
        for lst in [self.ints, self.floats, self.strings, self.objs]:
            if lst is not None and len(lst) > 0:
                it_list.append(lst)
        return it_list


@dataclass
class NumpyView:
    ints: Optional[np.ndarray] = None
    floats: Optional[np.ndarray] = None

    def iterables(self) -> List[np.ndarray]:
        it_list = []
        for lst in [self.ints, self.floats]:
            if lst is not None and len(lst) > 0:
                it_list.append(lst)
        return it_list

    @property
    def len(self) -> int:
        length = 0
        if self.ints is not None:
            length += len(self.ints)
        if self.floats is not None:
            length += len(self.floats)

        return length


@dataclass
class PandasView:
    strings: Optional[pd.Series] = None
    objs: Optional[pd.Series] = None

    def iterables(self) -> List[pd.Series]:
        it_list = []
        for lst in [self.strings, self.objs]:
            if lst is not None and len(lst) > 0:
                it_list.append(lst)
        return it_list


@dataclass(init=False)
class PreprocessedColumn:
    """View of a column with data of various underlying storage.

    If Pandas is available, we will use Pandas to handle batch processing.
    If numpy is available, we will use ndarray for numerical values.
    Otherwise, we preprocess values into typed lists for downstream consumers.
    We also track the null count and ensure that processed lists/Series don't contain null values.
    """

    numpy: NumpyView
    pandas: PandasView
    list: ListView
    # TODO: other data sources such as Apache Arrow here

    null_count: int = 0
    len: int = 0
    original: Any = None

    def __init__(self) -> None:
        self.numpy = NumpyView()
        self.pandas = PandasView()
        self.list = ListView()
        self.null_count = 0
        self.bool_count = 0
        self.bool_count_where_true = 0
        self.len = -1

    def _pandas_split(self, series: pd.Series, parse_numeric_string: bool = False) -> None:
        """
        Split a Pandas Series into numpy array and other Pandas series.

        Args:
            series: the original Pandas series
            parse_numeric_string: if set, this will coerce values into integer using pands.to_numeric() method.

        Returns:
            SplitSeries with multiple values, including numpy arrays for numbers, and strings as a Pandas Series.
        """
        if series is None:
            return None
        if pd.Series is None:
            return None

        self.null_count = len(series[series.isnull()])

        non_null_series = series[series.notnull()]
        if pdc.is_numeric_dtype(series.dtype) and not pdc.is_bool_dtype(series.dtype):
            if pdc.is_float_dtype(series.dtype):
                floats = non_null_series.astype(float)
                self.numpy.floats = floats
                return
            else:
                ints = non_null_series.astype(int)
                self.numpy.ints = ints
                return

        # if non_null_series is empty, then early exit.
        # this fixes a bug where empty columns produce masks of types other than bool
        # and DatetimeArrays do not support | operator for example.
        if non_null_series.empty:
            return

        if issubclass(series.dtype.type, str):
            self.pandas.strings = non_null_series
            return

        if parse_numeric_string:
            non_null_series = pd.to_numeric(non_null_series, errors="ignore")

        float_mask = non_null_series.apply(lambda x: pdc.is_float(x) or pdc.is_decimal(x))
        bool_mask = non_null_series.apply(lambda x: pdc.is_bool(x))
        bool_mask_where_true = non_null_series.apply(lambda x: pdc.is_bool(x) and x)
        int_mask = non_null_series.apply(lambda x: pdc.is_number(x) and pdc.is_integer(x) and not pdc.is_bool(x))
        str_mask = non_null_series.apply(lambda x: isinstance(x, str))

        floats = non_null_series[float_mask]
        ints = non_null_series[int_mask]
        bool_count = non_null_series[bool_mask].count()
        bool_count_where_true = non_null_series[bool_mask_where_true].count()
        strings = non_null_series[str_mask]
        objs = non_null_series[~(float_mask | str_mask | int_mask | bool_mask)]

        # convert numeric types to float if they are considered
        # Fractional types e.g. decimal.Decimal only if there are values
        if not floats.empty:
            floats = floats.astype(float)
        self.numpy = NumpyView(floats=floats, ints=ints)
        self.pandas.strings = strings
        self.pandas.objs = objs
        self.bool_count = bool_count
        self.bool_count_where_true = bool_count_where_true

    def raw_iterator(self) -> Iterator[Any]:
        iterables = [
            *self.numpy.iterables(),
            *self.pandas.iterables(),
            *self.list.iterables(),
        ]
        return itertools.chain(iterables)

    @staticmethod
    def apply(data: Any) -> "PreprocessedColumn":
        result = PreprocessedColumn()
        result.original = data
        if pd.Series is not None and isinstance(data, pd.Series):
            result._pandas_split(data)
            result.len = len(data)
            return result

        if isinstance(data, np.ndarray):
            result.len = len(data)
            if issubclass(data.dtype.type, np.number):
                if issubclass(data.dtype.type, np.floating):
                    result.numpy = NumpyView(floats=data.astype(float))
                else:
                    result.numpy = NumpyView(ints=data.astype(int))
            else:
                raise ValueError(f"Unsupported numpy type: {data.dtype}")
            return result

        if isinstance(data, List):
            result.len = len(data)
            if pd.Series:
                return PreprocessedColumn.apply(pd.Series(data, dtype="object"))

            int_list = []
            float_list = []
            string_list = []
            obj_list = []
            null_count = 0
            for x in data:
                if isinstance(x, int):
                    int_list.append(x)
                elif isinstance(x, float):
                    float_list.append(x)
                elif isinstance(x, str):
                    string_list.append(x)
                elif x is not None:
                    obj_list.append(x)
                else:
                    null_count += 1

            result.null_count = null_count
            if np.ndarray:
                ints = np.asarray(int_list, dtype=int)
                floats = np.asarray(float_list, dtype=float)

                result.numpy = NumpyView(ints=ints, floats=floats)
                result.list = ListView(strings=string_list, objs=obj_list)
                return result
            else:
                result.list = ListView(ints=int_list, floats=float_list, strings=string_list, objs=obj_list)
                return result

        if isinstance(data, Iterable) or isinstance(data, Iterator):
            logger.warning(
                "Materializing an Iterable or Iterator into a list for processing. This could cause memory issue"
            )
            list_format = list(data)
            return PreprocessedColumn.apply(list_format)

        # scalars - when processing dictionary entries
        logger.info(f"Warning single value passed as column data, wrapping type: {type(data)} in list")
        list_format = [data]
        return PreprocessedColumn.apply(list_format)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.any_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import google.protobuf.struct_pb2
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class DataType(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class _Type:
        ValueType = typing.NewType('ValueType', builtins.int)
        V: typing_extensions.TypeAlias = ValueType
    class _TypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[DataType._Type.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        UNKNOWN: DataType._Type.ValueType  # 0
        NULL: DataType._Type.ValueType  # 1
        FRACTIONAL: DataType._Type.ValueType  # 2
        INTEGRAL: DataType._Type.ValueType  # 3
        BOOLEAN: DataType._Type.ValueType  # 4
        STRING: DataType._Type.ValueType  # 5
    class Type(_Type, metaclass=_TypeEnumTypeWrapper):
        pass

    UNKNOWN: DataType.Type.ValueType  # 0
    NULL: DataType.Type.ValueType  # 1
    FRACTIONAL: DataType.Type.ValueType  # 2
    INTEGRAL: DataType.Type.ValueType  # 3
    BOOLEAN: DataType.Type.ValueType  # 4
    STRING: DataType.Type.ValueType  # 5

    TYPE_FIELD_NUMBER: builtins.int
    type: global___DataType.Type.ValueType
    def __init__(self,
        *,
        type: global___DataType.Type.ValueType = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["type",b"type"]) -> None: ...
global___DataType = DataType

class HllSketchMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    SKETCH_FIELD_NUMBER: builtins.int
    sketch: builtins.bytes
    def __init__(self,
        *,
        sketch: builtins.bytes = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["sketch",b"sketch"]) -> None: ...
global___HllSketchMessage = HllSketchMessage

class FrequentItemsSketchMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    SKETCH_FIELD_NUMBER: builtins.int
    sketch: builtins.bytes
    def __init__(self,
        *,
        sketch: builtins.bytes = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["sketch",b"sketch"]) -> None: ...
global___FrequentItemsSketchMessage = FrequentItemsSketchMessage

class KllSketchMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    SKETCH_FIELD_NUMBER: builtins.int
    sketch: builtins.bytes
    def __init__(self,
        *,
        sketch: builtins.bytes = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["sketch",b"sketch"]) -> None: ...
global___KllSketchMessage = KllSketchMessage

class CpcSketchMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    SKETCH_FIELD_NUMBER: builtins.int
    sketch: builtins.bytes
    def __init__(self,
        *,
        sketch: builtins.bytes = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["sketch",b"sketch"]) -> None: ...
global___CpcSketchMessage = CpcSketchMessage

class MetricComponentMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    TYPE_ID_FIELD_NUMBER: builtins.int
    N_FIELD_NUMBER: builtins.int
    D_FIELD_NUMBER: builtins.int
    FREQUENT_ITEMS_FIELD_NUMBER: builtins.int
    HLL_FIELD_NUMBER: builtins.int
    KLL_FIELD_NUMBER: builtins.int
    CPC_FIELD_NUMBER: builtins.int
    SERIALIZED_BYTES_FIELD_NUMBER: builtins.int
    DATACLASS_PARAM_FIELD_NUMBER: builtins.int
    MSG_FIELD_NUMBER: builtins.int
    type_id: builtins.int
    """type ID. It might determine how a field gets serder and the aggregation logic"""

    n: builtins.int
    """first class citizens"""

    d: builtins.float
    @property
    def frequent_items(self) -> global___FrequentItemsSketchMessage: ...
    @property
    def hll(self) -> global___HllSketchMessage: ...
    @property
    def kll(self) -> global___KllSketchMessage: ...
    @property
    def cpc(self) -> global___CpcSketchMessage: ...
    serialized_bytes: builtins.bytes
    """extension points"""

    @property
    def dataclass_param(self) -> google.protobuf.struct_pb2.Struct: ...
    @property
    def msg(self) -> google.protobuf.any_pb2.Any: ...
    def __init__(self,
        *,
        type_id: builtins.int = ...,
        n: builtins.int = ...,
        d: builtins.float = ...,
        frequent_items: typing.Optional[global___FrequentItemsSketchMessage] = ...,
        hll: typing.Optional[global___HllSketchMessage] = ...,
        kll: typing.Optional[global___KllSketchMessage] = ...,
        cpc: typing.Optional[global___CpcSketchMessage] = ...,
        serialized_bytes: builtins.bytes = ...,
        dataclass_param: typing.Optional[google.protobuf.struct_pb2.Struct] = ...,
        msg: typing.Optional[google.protobuf.any_pb2.Any] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["cpc",b"cpc","d",b"d","dataclass_param",b"dataclass_param","frequent_items",b"frequent_items","hll",b"hll","kll",b"kll","msg",b"msg","n",b"n","serialized_bytes",b"serialized_bytes","value",b"value"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["cpc",b"cpc","d",b"d","dataclass_param",b"dataclass_param","frequent_items",b"frequent_items","hll",b"hll","kll",b"kll","msg",b"msg","n",b"n","serialized_bytes",b"serialized_bytes","type_id",b"type_id","value",b"value"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["value",b"value"]) -> typing.Optional[typing_extensions.Literal["n","d","frequent_items","hll","kll","cpc","serialized_bytes","dataclass_param","msg"]]: ...
global___MetricComponentMessage = MetricComponentMessage

class MetricMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class MetricComponentsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text
        @property
        def value(self) -> global___MetricComponentMessage: ...
        def __init__(self,
            *,
            key: typing.Text = ...,
            value: typing.Optional[global___MetricComponentMessage] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value",b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    METRIC_COMPONENTS_FIELD_NUMBER: builtins.int
    @property
    def metric_components(self) -> google.protobuf.internal.containers.MessageMap[typing.Text, global___MetricComponentMessage]: ...
    def __init__(self,
        *,
        metric_components: typing.Optional[typing.Mapping[typing.Text, global___MetricComponentMessage]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["metric_components",b"metric_components"]) -> None: ...
global___MetricMessage = MetricMessage

class ColumnMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class MetricComponentsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text
        @property
        def value(self) -> global___MetricComponentMessage: ...
        def __init__(self,
            *,
            key: typing.Text = ...,
            value: typing.Optional[global___MetricComponentMessage] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value",b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    METRIC_COMPONENTS_FIELD_NUMBER: builtins.int
    @property
    def metric_components(self) -> google.protobuf.internal.containers.MessageMap[typing.Text, global___MetricComponentMessage]: ...
    def __init__(self,
        *,
        metric_components: typing.Optional[typing.Mapping[typing.Text, global___MetricComponentMessage]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["metric_components",b"metric_components"]) -> None: ...
global___ColumnMessage = ColumnMessage

class DatasetProperties(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class TagsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text
        value: typing.Text
        def __init__(self,
            *,
            key: typing.Text = ...,
            value: typing.Text = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    class MetadataEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text
        value: typing.Text
        def __init__(self,
            *,
            key: typing.Text = ...,
            value: typing.Text = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    SCHEMA_MAJOR_VERSION_FIELD_NUMBER: builtins.int
    SCHEMA_MINOR_VERSION_FIELD_NUMBER: builtins.int
    CREATION_TIMESTAMP_FIELD_NUMBER: builtins.int
    DATASET_TIMESTAMP_FIELD_NUMBER: builtins.int
    TAGS_FIELD_NUMBER: builtins.int
    METADATA_FIELD_NUMBER: builtins.int
    schema_major_version: builtins.int
    schema_minor_version: builtins.int
    creation_timestamp: builtins.int
    dataset_timestamp: builtins.int
    @property
    def tags(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, typing.Text]: ...
    @property
    def metadata(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, typing.Text]: ...
    def __init__(self,
        *,
        schema_major_version: builtins.int = ...,
        schema_minor_version: builtins.int = ...,
        creation_timestamp: builtins.int = ...,
        dataset_timestamp: builtins.int = ...,
        tags: typing.Optional[typing.Mapping[typing.Text, typing.Text]] = ...,
        metadata: typing.Optional[typing.Mapping[typing.Text, typing.Text]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["creation_timestamp",b"creation_timestamp","dataset_timestamp",b"dataset_timestamp","metadata",b"metadata","schema_major_version",b"schema_major_version","schema_minor_version",b"schema_minor_version","tags",b"tags"]) -> None: ...
global___DatasetProperties = DatasetProperties

class ChunkOffsets(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    OFFSETS_FIELD_NUMBER: builtins.int
    @property
    def offsets(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    def __init__(self,
        *,
        offsets: typing.Optional[typing.Iterable[builtins.int]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["offsets",b"offsets"]) -> None: ...
global___ChunkOffsets = ChunkOffsets

class ChunkMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class MetricComponentsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.int
        @property
        def value(self) -> global___MetricComponentMessage: ...
        def __init__(self,
            *,
            key: builtins.int = ...,
            value: typing.Optional[global___MetricComponentMessage] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value",b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    METRIC_COMPONENTS_FIELD_NUMBER: builtins.int
    @property
    def metric_components(self) -> google.protobuf.internal.containers.MessageMap[builtins.int, global___MetricComponentMessage]: ...
    def __init__(self,
        *,
        metric_components: typing.Optional[typing.Mapping[builtins.int, global___MetricComponentMessage]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["metric_components",b"metric_components"]) -> None: ...
global___ChunkMessage = ChunkMessage

class ChunkHeader(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class _ChunkType:
        ValueType = typing.NewType('ValueType', builtins.int)
        V: typing_extensions.TypeAlias = ValueType
    class _ChunkTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[ChunkHeader._ChunkType.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        DATASET: ChunkHeader._ChunkType.ValueType  # 0
        COLUMN: ChunkHeader._ChunkType.ValueType  # 2
    class ChunkType(_ChunkType, metaclass=_ChunkTypeEnumTypeWrapper):
        pass

    DATASET: ChunkHeader.ChunkType.ValueType  # 0
    COLUMN: ChunkHeader.ChunkType.ValueType  # 2

    TYPE_FIELD_NUMBER: builtins.int
    TAG_FIELD_NUMBER: builtins.int
    LENGTH_FIELD_NUMBER: builtins.int
    type: global___ChunkHeader.ChunkType.ValueType
    tag: typing.Text
    length: builtins.int
    def __init__(self,
        *,
        type: global___ChunkHeader.ChunkType.ValueType = ...,
        tag: typing.Text = ...,
        length: builtins.int = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["length",b"length","tag",b"tag","type",b"type"]) -> None: ...
global___ChunkHeader = ChunkHeader

class DatasetProfileHeader(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class ColumnOffsetsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text
        @property
        def value(self) -> global___ChunkOffsets: ...
        def __init__(self,
            *,
            key: typing.Text = ...,
            value: typing.Optional[global___ChunkOffsets] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value",b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    class IndexedMetricPathsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.int
        value: typing.Text
        def __init__(self,
            *,
            key: builtins.int = ...,
            value: typing.Text = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    PROPERTIES_FIELD_NUMBER: builtins.int
    COLUMN_OFFSETS_FIELD_NUMBER: builtins.int
    METRIC_OFFSETS_FIELD_NUMBER: builtins.int
    LENGTH_FIELD_NUMBER: builtins.int
    INDEXED_METRIC_PATHS_FIELD_NUMBER: builtins.int
    @property
    def properties(self) -> global___DatasetProperties: ...
    @property
    def column_offsets(self) -> google.protobuf.internal.containers.MessageMap[typing.Text, global___ChunkOffsets]: ...
    @property
    def metric_offsets(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ChunkOffsets]: ...
    length: builtins.int
    """total length of the rest of the data for this profile."""

    @property
    def indexed_metric_paths(self) -> google.protobuf.internal.containers.ScalarMap[builtins.int, typing.Text]: ...
    def __init__(self,
        *,
        properties: typing.Optional[global___DatasetProperties] = ...,
        column_offsets: typing.Optional[typing.Mapping[typing.Text, global___ChunkOffsets]] = ...,
        metric_offsets: typing.Optional[typing.Iterable[global___ChunkOffsets]] = ...,
        length: builtins.int = ...,
        indexed_metric_paths: typing.Optional[typing.Mapping[builtins.int, typing.Text]] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["properties",b"properties"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["column_offsets",b"column_offsets","indexed_metric_paths",b"indexed_metric_paths","length",b"length","metric_offsets",b"metric_offsets","properties",b"properties"]) -> None: ...
global___DatasetProfileHeader = DatasetProfileHeader

class SegmentTag(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    KEY_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    key: typing.Text
    value: typing.Text
    def __init__(self,
        *,
        key: typing.Text = ...,
        value: typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...
global___SegmentTag = SegmentTag

class Segment(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    TAGS_FIELD_NUMBER: builtins.int
    PARTITION_ID_FIELD_NUMBER: builtins.int
    @property
    def tags(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SegmentTag]: ...
    partition_id: typing.Text
    def __init__(self,
        *,
        tags: typing.Optional[typing.Iterable[global___SegmentTag]] = ...,
        partition_id: typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["partition_id",b"partition_id","tags",b"tags"]) -> None: ...
global___Segment = Segment

class SegmentationPartition(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    NAME_FIELD_NUMBER: builtins.int
    ID_FIELD_NUMBER: builtins.int
    name: typing.Text
    id: typing.Text
    def __init__(self,
        *,
        name: typing.Text = ...,
        id: typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["id",b"id","name",b"name"]) -> None: ...
global___SegmentationPartition = SegmentationPartition

class DatasetSegmentHeader(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class OffsetsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.int
        value: builtins.int
        def __init__(self,
            *,
            key: builtins.int = ...,
            value: builtins.int = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    HAS_SEGMENTS_FIELD_NUMBER: builtins.int
    SEGMENTS_FIELD_NUMBER: builtins.int
    OFFSETS_FIELD_NUMBER: builtins.int
    has_segments: builtins.bool
    """whether the file contains segmented data or not"""

    @property
    def segments(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Segment]:
        """list of segments"""
        pass
    @property
    def offsets(self) -> google.protobuf.internal.containers.ScalarMap[builtins.int, builtins.int]:
        """we want the indices of the map of offsets to have static size, and thus we use fixed32 bytes"""
        pass
    def __init__(self,
        *,
        has_segments: builtins.bool = ...,
        segments: typing.Optional[typing.Iterable[global___Segment]] = ...,
        offsets: typing.Optional[typing.Mapping[builtins.int, builtins.int]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["has_segments",b"has_segments","offsets",b"offsets","segments",b"segments"]) -> None: ...
global___DatasetSegmentHeader = DatasetSegmentHeader



from abc import ABC, abstractmethod
from types import FunctionType, LambdaType, MethodType
from typing import Any, Dict, Type, TypeVar, Union

DF = TypeVar("DF", bound="DecoratedFunction")

FuncType = Union[FunctionType, MethodType, LambdaType]


class DecoratedFunction(ABC):
    _func: FuncType  # noqa
    _name: str

    @classmethod
    @abstractmethod
    def build(cls: Type[DF], func: FuncType, name: str) -> DF:  # noqa
        pass


def _decorate_func(  # type: ignore
    *,
    key: Any,  # type: ignore
    name: str,  # type: ignore
    wrapper_dict: Dict[Any, DF],
    clazz: Type[DF],
):
    def decorated(func: FunctionType) -> DF:  # noqa
        return _func_wrapper(func=func, key=key, name=name, wrapper_dict=wrapper_dict, clazz=clazz)

    return decorated


def _func_wrapper(
    *,
    func: FuncType,
    key: Any,
    name: str,
    wrapper_dict: Dict[Any, DF],
    clazz: Type[DF],
) -> DF:
    existing_ = wrapper_dict.get(key)
    if existing_ is not None:
        # same annotation could be executed twice an import so we
        # want to detect duplication here. Probably buggy because we rely on
        # the code and name to detect duplicates? Good enough?
        if (
            existing_._func.__name__ == func.__name__
            and existing_._func.__code__.__hash__() == func.__code__.__hash__()  # type: ignore
        ):
            return existing_
        raise ValueError(f"Builtin aggregator for type {key} is already registered")

    wrapped_func = clazz.build(func=func, name=name)
    wrapper_dict[key] = wrapped_func
    return wrapped_func



from typing import Union

from whylogs.core.constraints import MetricConstraint, MetricsSelector
from whylogs.core.metrics import DistributionMetric


def greater_than_number(column_name: str, number: Union[float, int], skip_missing: bool = True) -> MetricConstraint:
    """Minimum value of given column must be above defined number.

    Parameters
    ----------
    column_name : str
        Column the constraint is applied to
    number : float
        reference value for applying the constraint
    skip_missing: bool
        If skip_missing is True, missing distribution metrics will make the check pass.
        If False, the check will fail on missing metrics
    """

    def is_greater(metric: DistributionMetric) -> bool:
        if not metric.kll.value.is_empty():
            return metric.min >= number
        else:
            return True if skip_missing else False

    constraint = MetricConstraint(
        name=f"{column_name} greater than number {number}",
        condition=is_greater,
        metric_selector=MetricsSelector(column_name=column_name, metric_name="distribution"),
    )
    return constraint


def smaller_than_number(column_name: str, number: float, skip_missing: bool = True) -> MetricConstraint:
    """Maximum value of given column must be below defined number.

    Parameters
    ----------
    column_name : str
        Column the constraint is applied to
    number : float
        reference value for applying the constraint
    skip_missing: bool
        If skip_missing is True, missing distribution metrics will make the check pass.
        If False, the check will fail on missing metrics
    """

    def is_smaller(metric) -> bool:
        if not metric.kll.value.is_empty():
            return number >= metric.max
        else:
            return True if skip_missing else False

    constraint = MetricConstraint(
        name=f"{column_name} smaller than number {number}",
        condition=is_smaller,
        metric_selector=MetricsSelector(column_name=column_name, metric_name="distribution"),
    )
    return constraint


def mean_between_range(column_name: str, lower: float, upper: float, skip_missing: bool = True) -> MetricConstraint:
    """Estimated mean must be between range defined by lower and upper bounds.

    Parameters
    ----------
    column_name : str
        Column the constraint is applied to
    lower : int
        Lower bound of defined range
    upper : int
        Upper bound of the value range
    skip_missing: bool
        If skip_missing is True, missing distribution metrics will make the check pass.
        If False, the check will fail on missing metrics
    """

    def is_mean_between(metric) -> bool:
        if not metric.kll.value.is_empty():
            return lower <= metric.avg <= upper
        else:
            return True if skip_missing else False

    constraint = MetricConstraint(
        name=f"{column_name} mean between {lower} and {upper} (inclusive)",
        condition=is_mean_between,
        metric_selector=MetricsSelector(column_name=column_name, metric_name="distribution"),
    )
    return constraint


def stddev_between_range(column_name: str, lower: float, upper: float, skip_missing: bool = True):
    """Estimated standard deviation must be between range defined by lower and upper bounds.

    Parameters
    ----------
    column_name: str
        Column the constraint is applied to
    lower: float
        Lower bound of defined range
    upper: float
        Upper bound of the value range
    skip_missing: bool
        If skip_missing is True, missing distribution metrics will make the check pass.
        If False, the check will fail on missing metrics
    """

    def is_stddev_between_range(metric):
        if not metric.kll.value.is_empty():
            return lower <= metric.stddev <= upper
        else:
            return True if skip_missing else False

    constraint = MetricConstraint(
        name=f"{column_name} standard deviation between {lower} and {upper} (inclusive)",
        condition=is_stddev_between_range,
        metric_selector=MetricsSelector(column_name=column_name, metric_name="distribution"),
    )
    return constraint


def quantile_between_range(column_name: str, quantile: float, lower: float, upper: float, skip_missing: bool = True):
    """Q-th quantile value must be withing the range defined by lower and upper boundaries.

    Parameters
    ----------
    column_name: str
        Column the constraint is applied to
    quantile: float
        Quantile value. E.g. median is equal to quantile_value=0.5
    lower: float
        Lower bound of defined range
    upper: float
        Upper bound of the value range
    skip_missing: bool
        If skip_missing is True, missing distribution metrics will make the check pass.
        If False, the check will fail on missing metrics
    """

    def quantile_in_range(metric):
        if not metric.kll.value.is_empty():
            quantile_value = metric.kll.value.get_quantile(quantile)
            result: bool = lower <= quantile_value <= upper
            return result
        else:
            return True if skip_missing else False

    constraint = MetricConstraint(
        name=f"{column_name} {quantile}-th quantile value between {lower} and {upper} (inclusive)",
        condition=quantile_in_range,
        metric_selector=MetricsSelector(column_name=column_name, metric_name="distribution"),
    )
    return constraint



from .profiler import (
    collect_column_profile_views,
    collect_dataset_profile_view,
    column_profile_bytes_aggregator,
    whylogs_pandas_map_profiler,
)

__ALL__ = [
    whylogs_pandas_map_profiler,
    column_profile_bytes_aggregator,
    collect_column_profile_views,
    collect_dataset_profile_view,
]



import logging
from copy import deepcopy
from dataclasses import dataclass, field
from typing import Any, Dict, List, Mapping, Optional, TypeVar

from whylogs.core.datatypes import StandardTypeMapper, TypeMapper
from whylogs.core.metrics.metrics import Metric, MetricConfig
from whylogs.core.resolvers import Resolver, StandardResolver
from whylogs.core.segmentation_partition import SegmentationPartition
from whylogs.core.stubs import pd
from whylogs.core.validators.validator import Validator

logger = logging.getLogger(__name__)

LARGE_CACHE_SIZE_LIMIT = 1024 * 100

T = TypeVar("T", bound="DatasetSchema")


class DatasetSchema:
    """
    Defines the schema for tracking metrics in whylogs.

    In order to customize your tracking, you can extend this class to specify your own column schema
    or your own type resolution. Otherwise, you can just use the default DatasetSchema object.

    Schema objects are also used to group datasets together.

    Attributes:
        types: required. a dictionay of column name to the Python type.

        default_configs: optional. Options to configure various behavior of whylogs.

        type_mapper: Optional. a mapper that transates the Python type to standardized whylogs :class:`DataType` object.

        resolvers: Optional. an object that defines how to map from a column name, a whylogs :class:`DataType` and a
                   schema to metrics.


    Examples
    --------
    >>> import pandas as pd
    >>> import numpy as np
    >>> from whylogs.core import DatasetSchema, DatasetProfile
    >>> from whylogs.core.resolvers import Resolver, StandardResolver
    >>>
    >>> class MyResolver(StandardResolver):
    ...    pass
    >>>
    >>> schema = DatasetSchema(
    ...    types={
    ...        "col1": str,
    ...        "col2": np.int32,
    ...        "col3": pd.CategoricalDtype(categories=('foo', 'bar'), ordered=True)
    ...    },
    ...    resolvers=MyResolver()
    ... )
    >>> prof = DatasetProfile(schema)
    >>> df = pd.DataFrame({"col1": ['foo'], "col2": np.array([1], dtype=np.int32), "col3": ['bar']})
    >>> prof.track(pandas=df)

    """

    def __init__(
        self,
        types: Optional[Dict[str, Any]] = None,
        default_configs: Optional[MetricConfig] = None,
        type_mapper: Optional[TypeMapper] = None,
        resolvers: Optional[Resolver] = None,
        cache_size: int = 1024,
        schema_based_automerge: bool = False,
        segments: Optional[Dict[str, SegmentationPartition]] = None,
        validators: Dict[str, List[Validator]] = None,
    ) -> None:
        self._columns = dict()
        self.types = types or dict()
        self.default_configs = default_configs or MetricConfig()
        self.type_mapper = type_mapper or StandardTypeMapper()
        self.resolvers = resolvers or StandardResolver()
        self.cache_size = cache_size
        self.schema_based_automerge = schema_based_automerge
        self.segments = segments or dict()
        self.validators = validators or dict()

        if self.cache_size < 0:
            logger.warning("Negative cache size value. Disabling caching")
            self.cache_size = 0

        if self.cache_size > LARGE_CACHE_SIZE_LIMIT:
            logger.warning(
                "Cache size value %s is larger than recommended limit of %s",
                self.cache_size,
                LARGE_CACHE_SIZE_LIMIT,
            )

        for col, data_type in self.types.items():
            self._columns[col] = ColumnSchema(
                dtype=data_type,
                resolver=self.resolvers,
                type_mapper=self.type_mapper,
                cfg=self.default_configs,
                validators=self.validators,
            )

    def copy(self) -> "DatasetSchema":
        """Returns a new instance of the same underlying schema"""
        key_dict = vars(self).copy()
        key_dict.pop("_columns")
        keys = key_dict.keys()
        args = {k: deepcopy(self.__dict__[k]) for k in keys if k not in self.types}
        copy = self.__class__(**args)
        copy._columns = deepcopy(self._columns)
        copy.segments = self.segments.copy()
        return copy

    def resolve(self, *, pandas: Optional[pd.DataFrame] = None, row: Optional[Mapping[str, Any]] = None) -> bool:
        if pandas is not None:
            return self._resolve_pdf(pandas)

        if row is not None:
            for k, v in row.items():
                if k in self._columns:
                    continue

                self._columns[k] = ColumnSchema(
                    dtype=type(v),
                    cfg=self.default_configs,
                    resolver=self.resolvers,
                    validators=self.validators,
                    type_mapper=self.type_mapper,
                )
            return True
        raise NotImplementedError

    def _resolve_pdf(self, df: pd.DataFrame, force_resolve: bool = False) -> bool:
        """
        Resolve ColumnSchema from the dataframe. We only resolve newly detected
        columns unless `force_resolve` is set to True.
        """
        col_names = df.dtypes.keys()
        dirty = False
        for col_name in col_names:
            if not force_resolve and col_name in self._columns:
                logger.debug("column is already configured")
                continue

            col_dtype = df.dtypes[col_name]
            self._columns[col_name] = ColumnSchema(
                dtype=col_dtype,
                cfg=self.default_configs,
                resolver=self.resolvers,
                validators=self.validators,
                type_mapper=self.type_mapper,
            )
            dirty = True

        return dirty

    def get_col_names(self) -> tuple:
        return tuple(self._columns.keys())

    def get(self, name: str) -> Optional["ColumnSchema"]:
        return self._columns.get(name)


@dataclass(frozen=True)
class ColumnSchema:
    """
    Schema of a column.

    The main goal is to specify the data type.
    On top of that, users can configure their own tracker resolution logic (mapping a type to a list of tracker
    factories) and any additional trackers here.
    """

    dtype: Any
    cfg: MetricConfig = MetricConfig()
    type_mapper: TypeMapper = StandardTypeMapper()
    resolver: Resolver = StandardResolver()
    validators: Dict[str, List[Validator]] = field(default_factory=dict)

    def get_metrics(self, name: str) -> Dict[str, Metric]:
        return self.resolver.resolve(name=name, why_type=self.type_mapper(self.dtype), column_schema=self)

    def get_validators(self, name: str) -> List[Optional[Validator]]:
        if self.validators:
            return self.validators.get(name, [])
        return []



import json
from logging import getLogger
from typing import Any, Dict, List, Optional, Union

from typing_extensions import TypedDict

from whylogs.core.configs import SummaryConfig
from whylogs.core.metrics import CardinalityMetric
from whylogs.core.utils import get_distribution_metrics
from whylogs.core.view.column_profile_view import ColumnProfileView
from whylogs.core.view.dataset_profile_view import DatasetProfileView
from whylogs.viz.utils import (
    DescriptiveStatistics,
    QuantileStats,
    _calculate_descriptive_statistics,
    _calculate_quantile_statistics,
)
from whylogs.viz.utils.descriptive_stats import _get_count_metrics_from_column_view
from whylogs.viz.utils.drift_calculations import calculate_drift_values
from whylogs.viz.utils.frequent_items_calculations import (
    FrequentItemEstimate,
    frequent_items_from_view,
)
from whylogs.viz.utils.histogram_calculations import (
    HistogramSummary,
    histogram_from_view,
)

logger = getLogger(__name__)


class FeatureStats(TypedDict):
    total_count: Optional[float]
    missing: Optional[float]
    distinct: Optional[float]
    min: Optional[float]
    max: Optional[float]
    range: Optional[float]
    quantile_statistics: Optional[QuantileStats]
    descriptive_statistics: Optional[DescriptiveStatistics]


def _get_cardinality_metrics_from_column_view(
    column_view: ColumnProfileView, count_n: Optional[float] = None, count_missing: Optional[float] = None
) -> Union[None, float]:
    cardinality: Optional[CardinalityMetric] = column_view.get_metric("cardinality")
    if cardinality is not None and count_n is not None and count_missing is not None:
        card_estimate = cardinality.hll.value.get_estimate()
        if (count_n - count_missing) != 0:
            distinct = card_estimate / (count_n - count_missing) * 100
            return distinct
    return None


def add_feature_statistics(
    feature_name: str,
    column_view: Union[None, ColumnProfileView],
) -> Dict[str, FeatureStats]:
    if column_view is None:
        return {}

    feature_with_statistics: Dict[str, FeatureStats] = {}

    total_count, count_n, count_missing = _get_count_metrics_from_column_view(column_view=column_view)
    distinct = _get_cardinality_metrics_from_column_view(
        column_view=column_view, count_n=count_n, count_missing=count_missing
    )
    min_val, max_val, range_val = get_distribution_metrics(column_view=column_view)

    quantile_stats = _calculate_quantile_statistics(column_view)
    desc_stats = _calculate_descriptive_statistics(column_view)

    feature_stats: FeatureStats = {
        "total_count": total_count,
        "missing": count_missing,
        "distinct": distinct,
        "min": min_val,
        "max": max_val,
        "range": range_val,
        "quantile_statistics": quantile_stats,
        "descriptive_statistics": desc_stats,
    }
    feature_with_statistics[feature_name] = feature_stats

    return feature_with_statistics


class OverallStats(TypedDict):
    observations: int
    missing_cells: int
    missing_percentage: float


class ColumnSummary(TypedDict):
    histogram: Optional[HistogramSummary]
    frequentItems: Optional[List[FrequentItemEstimate]]
    drift_from_ref: Optional[float]
    isDiscrete: Optional[bool]
    featureStats: Optional[FeatureStats]


class DatasetSummary(TypedDict):
    columns: Dict[str, Optional[ColumnSummary]]
    properties: Optional[OverallStats]


def generate_profile_summary(
    target_view: DatasetProfileView, config: Optional[SummaryConfig]
) -> Optional[Dict[str, Any]]:
    if config is None:
        config = SummaryConfig()

    if not target_view:
        raise ValueError("This method has to get target Dataset Profile View")

    overall_stats: OverallStats = add_overall_statistics(target_view)
    target_col_views = target_view.get_columns()
    target_summary: DatasetSummary = {"columns": {}, "properties": overall_stats}

    for target_col_name in target_col_views:
        target_column_summary: ColumnSummary = {
            "histogram": None,
            "frequentItems": None,
            "drift_from_ref": None,
            "isDiscrete": None,
            "featureStats": None,
        }

        target_col_view = target_col_views[target_col_name]
        target_stats = add_feature_statistics(target_col_name, target_col_view)
        target_column_summary["featureStats"] = target_stats[target_col_name]
        target_dist = target_col_view.get_metric("distribution")

        if target_dist and not target_dist.kll.value.is_empty():
            target_column_summary["isDiscrete"] = False
            target_histogram = histogram_from_view(target_col_view, target_col_name)
            target_column_summary["histogram"] = target_histogram
        elif target_col_view.get_metric("frequent_items"):
            target_column_summary["isDiscrete"] = True
            target_frequent_items = frequent_items_from_view(target_col_view, target_col_name, config)
            target_column_summary["frequentItems"] = target_frequent_items

        target_summary["columns"][target_col_name] = target_column_summary

    summaries = {
        "profile_from_whylogs": json.dumps(target_summary),
    }
    return summaries


def generate_summaries(
    target_view: DatasetProfileView, ref_view: Optional[DatasetProfileView], config: Optional[SummaryConfig]
) -> Optional[Dict[str, Any]]:
    if config is None:
        config = SummaryConfig()

    if not target_view or not ref_view:
        raise ValueError("This method has to get both target and reference profiles")

    overall_stats: OverallStats = add_overall_statistics(target_view)
    drift_values = calculate_drift_values(target_view, ref_view)

    target_col_views = target_view.get_columns()
    ref_col_views = ref_view.get_columns()
    ref_summary: DatasetSummary = {"columns": {}, "properties": overall_stats}
    target_summary: DatasetSummary = {"columns": {}, "properties": None}
    for target_col_name in target_col_views:
        if target_col_name in ref_col_views:
            target_column_summary: ColumnSummary = {
                "histogram": None,
                "frequentItems": None,
                "drift_from_ref": None,
                "isDiscrete": None,
                "featureStats": None,
            }
            ref_column_summary: ColumnSummary = {
                "histogram": None,
                "frequentItems": None,
                "drift_from_ref": None,
                "isDiscrete": None,
                "featureStats": None,
            }

            target_col_view = target_col_views[target_col_name]
            ref_col_view = ref_col_views[target_col_name]
            if not target_col_view or not ref_col_view:
                continue

            target_stats = add_feature_statistics(target_col_name, target_col_view)
            target_column_summary["featureStats"] = target_stats[target_col_name]

            if target_col_name in drift_values:
                col_drift_value = drift_values[target_col_name]
                if col_drift_value:
                    ref_column_summary["drift_from_ref"] = col_drift_value["p_value"]
            target_dist = target_col_view.get_metric("distribution")
            reference_dist = ref_col_view.get_metric("distribution")
            if (
                target_dist
                and reference_dist
                and not target_dist.kll.value.is_empty()
                and not reference_dist.kll.value.is_empty()
            ):
                target_column_summary["isDiscrete"] = ref_column_summary["isDiscrete"] = False

                target_histogram = histogram_from_view(target_col_view, target_col_name)
                target_column_summary["histogram"] = target_histogram

                ref_histogram = histogram_from_view(ref_col_view, target_col_name)
                ref_column_summary["histogram"] = ref_histogram

            elif target_col_view.get_metric("frequent_items") and ref_col_view.get_metric("frequent_items"):
                target_column_summary["isDiscrete"] = ref_column_summary["isDiscrete"] = True

                target_frequent_items = frequent_items_from_view(target_col_view, target_col_name, config)
                target_column_summary["frequentItems"] = target_frequent_items

                ref_frequent_items = frequent_items_from_view(ref_col_view, target_col_name, config)
                ref_column_summary["frequentItems"] = ref_frequent_items
            target_summary["columns"][target_col_name] = target_column_summary
            ref_summary["columns"][target_col_name] = ref_column_summary

    summaries = {
        "profile_from_whylogs": json.dumps(target_summary),
        "reference_profile_from_whylogs": json.dumps(ref_summary),
    }
    return summaries


def add_overall_statistics(target_view: DatasetProfileView) -> OverallStats:
    observations: int = 0
    missing_cells: int = 0
    missing_percentage: float = 0
    target_col_views = target_view.get_columns()
    for target_col_view in target_col_views.values():
        if target_col_view:
            cnt_metric = target_col_view.get_metric("counts")
            if cnt_metric:
                observations += cnt_metric.n.value
                null_count = cnt_metric.null.value
                missing_cells += null_count if null_count else 0
    missing_percentage = (missing_cells / observations) * 100 if observations else 0
    overall_stats: OverallStats = {
        "observations": observations,
        "missing_cells": missing_cells,
        "missing_percentage": missing_percentage,
    }
    return overall_stats



class Error(Exception):
    """Base error type for this module."""


class DeserializationError(Error):
    """Exception raised when deserializing data."""


class SerializationError(Error):
    """Exception raised when serializing data."""


class UnsupportedError(Error):
    """Exception raised when an operation is not supported."""


class LoggingError(Error):
    """Exception when an error happens during logging."""


class BadConfigError(Error):
    """Exception when an error is due to bad configuration."""



from datetime import datetime
from functools import reduce
from logging import getLogger
from typing import Dict, Iterable, Optional, Tuple

import whylogs as why
from whylogs.api.usage_stats import emit_usage
from whylogs.core.stubs import pd
from whylogs.core.view.column_profile_view import ColumnProfileView
from whylogs.core.view.dataset_profile_view import DatasetProfileView

logger = getLogger(__name__)
emit_usage("pyspark")

try:  # type: ignore
    from pyspark.sql import DataFrame as SparkDataFrame  # types: ignore
except ImportError:  # noqa
    logger.error("No pyspark available")
    SparkDataFrame = None  # type: ignore

COL_NAME_FIELD = "col_name"

COL_PROFILE_FIELD = "col_profile"


def whylogs_pandas_map_profiler(pdf_iterator: Iterable[pd.DataFrame]) -> Iterable[pd.DataFrame]:
    for input_df in pdf_iterator:
        res = why.log(input_df)
        res_df = pd.DataFrame(columns=[COL_NAME_FIELD, COL_PROFILE_FIELD])
        for col_name, col_profile in res.view().get_columns().items():
            d = {
                COL_NAME_FIELD: [col_name],
                COL_PROFILE_FIELD: [col_profile.to_protobuf().SerializeToString()],
            }
            df_temp = pd.DataFrame(data=d)
            res_df = res_df.append(df_temp)
        yield res_df


def column_profile_bytes_aggregator(group_by_cols: Tuple[str], profiles_df: pd.DataFrame) -> pd.DataFrame:
    if len(profiles_df) == 0:
        return pd.DataFrame([group_by_cols + (bytearray(),)])

    merged_profile: ColumnProfileView = reduce(
        lambda acc, x: acc.merge(x), profiles_df[COL_PROFILE_FIELD].apply(ColumnProfileView.from_bytes)
    )
    return pd.DataFrame([group_by_cols + (merged_profile.serialize(),)])


def collect_column_profile_views(input_df: SparkDataFrame) -> Dict[str, ColumnProfileView]:
    if SparkDataFrame is None:
        logger.warning("Unable to load pyspark; install pyspark to get whylogs profiling support in spark environment.")

    cp = f"{COL_NAME_FIELD} string, {COL_PROFILE_FIELD} binary"
    profile_bytes_df = input_df.mapInPandas(whylogs_pandas_map_profiler, schema=cp)  # type: ignore
    column_profiles = profile_bytes_df.groupby(COL_NAME_FIELD).applyInPandas(  # linebreak
        column_profile_bytes_aggregator, schema=cp
    )
    collected_profile_views = {
        row.col_name: ColumnProfileView.from_bytes(row.col_profile) for row in column_profiles.collect()
    }
    return collected_profile_views


def collect_dataset_profile_view(
    input_df: SparkDataFrame, dataset_timestamp: Optional[int] = None, creation_timestamp: Optional[int] = None
) -> DatasetProfileView:
    now = datetime.utcnow()

    _dataset_timestamp = dataset_timestamp or now
    _creation_timestamp = creation_timestamp or now

    column_views_dict = collect_column_profile_views(input_df=input_df)

    profile_view = DatasetProfileView(
        columns=column_views_dict, dataset_timestamp=_dataset_timestamp, creation_timestamp=_creation_timestamp
    )

    return profile_view



import logging
import tempfile
from typing import Any, Optional

import boto3
from botocore.client import BaseClient
from botocore.exceptions import ClientError

from whylogs.api.writer import Writer
from whylogs.api.writer.writer import Writable
from whylogs.core.utils import deprecated_alias

logger = logging.getLogger(__name__)


class S3Writer(Writer):
    """
    A WhyLogs writer to upload DatasetProfileViews onto Amazon S3.

    >**IMPORTANT**: In order to correctly connect to your Amazon S3 bucket, make sure you have
    the following environment variables set: `[AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY]`

    Parameters
    ----------
    s3_client: BaseClient, optional
        The s3 client used to authenticate and perform operations on the s3 bucket.
        Should be a BaseClient from the boto3 library
    base_prefix: str, optional
        The base file prefix for s3, in order to organize. A placeholder 'profile' will take place if None is provided.
    bucket_name: str, optional
        The name of the bucket to connect to. Made optional so the user can also access it
        via the `option` method
    object_name: str, optional
        The s3's object name. It basically states the location where the file goes to.
        Also made optional, so it can be defined through the `option` method
    Returns
    -------
        None

    Examples
    --------
    An example usage of this method can be represented with the simple code above. Here we
    are assuming the user has already assigned a pandas.DataFrame to the `df` variable.

    ```python
    import whylogs as why

    profile = why.log(pandas=df)
    profile.writer("s3").option(bucket_name="my_bucket").write()
    ```

    """

    def __init__(
        self,
        s3_client: Optional[BaseClient] = None,
        base_prefix: Optional[str] = None,
        bucket_name: Optional[str] = None,
        object_name: Optional[str] = None,
    ):
        self.s3_client = s3_client or boto3.client("s3")
        self.base_prefix = base_prefix or "profile"
        self.bucket_name = bucket_name or ""
        self.object_name = object_name or None

    @deprecated_alias(profile="file")
    def write(
        self,
        file: Writable,
        dest: Optional[str] = None,
        **kwargs: Any,
    ) -> None:

        dest = dest or file.get_default_path()  # type: ignore
        if self.object_name is None:
            self.object_name = dest
        try:
            with tempfile.NamedTemporaryFile() as tmp_file:
                file.write(path=tmp_file.name)  # type: ignore
                tmp_file.flush()
                self.s3_client.upload_file(tmp_file.name, self.bucket_name, self.object_name)
        except ClientError as e:
            logging.error(e)

    def option(
        self,
        bucket_name: Optional[str] = None,
        object_name: Optional[str] = None,
        s3_client: Optional[BaseClient] = None,
    ) -> None:  # type: ignore
        if bucket_name:
            self.bucket_name = bucket_name
        if object_name:
            self.object_name = object_name
        if s3_client:
            self.s3_client = s3_client



"""
Auto-generated protobuf class definitions.

Protobuf allows us to serialize/deserialize classes across languages
"""
from .v0_constraints_pb2 import *  # noqa
from .v0_messages_pb2 import *  # noqa
from .v0_summaries_pb2 import *  # noqa



"""
whylogs is an open source library for logging any kind of data. With whylogs, users are able to generate summaries
of their datasets (called whylogs profiles) which they can use to:

* Track changes in their dataset
* Create data constraints to know whether their data looks they way it should
* Quickly visualize key summary statistics about their datasets

These three functionalities enable a variety of use cases for data scientists, machine learning engineers, and
 data engineers:

* Detecting data drift (and resultant ML model performance degradation)
* Data quality validation
* Exploratory data analysis via data profiling
* Tracking data for ML experiments
* And many more...
"""
from .api import (
    ResultSet,
    log,
    log_classification_metrics,
    log_regression_metrics,
    logger,
    profiling,
    read,
    reader,
    write,
)
from .api.usage_stats import emit_usage as __emit_usage_stats
from .core import DatasetProfileView
from .migration.converters import v0_to_v1_view


def package_version(package: str = __package__) -> str:
    """Calculate version number based on pyproject.toml"""
    try:
        from importlib import metadata  # type: ignore
    except ImportError:
        # Python < 3.8
        import importlib_metadata as metadata  # type: ignore

    try:
        version = metadata.version(package)
    except metadata.PackageNotFoundError:  # type: ignore
        version = f"{package} is not installed."

    return version


__version__ = package_version()

__ALL__ = [
    ResultSet,
    DatasetProfileView,
    log,
    log_classification_metrics,
    log_regression_metrics,
    logger,
    read,
    reader,
    write,
    profiling,
    v0_to_v1_view,
    __version__,
]

__emit_usage_stats("import")



import datetime
import logging
import os
import tempfile
from typing import Any, Dict, Optional

import requests  # type: ignore
import whylabs_client  # type: ignore
from whylabs_client import ApiClient, Configuration  # type: ignore
from whylabs_client.api.log_api import AsyncLogResponse  # type: ignore
from whylabs_client.api.log_api import (
    LogApi,
    LogAsyncRequest,
    LogReferenceRequest,
    LogReferenceResponse,
)
from whylabs_client.rest import ForbiddenException  # type: ignore

from whylogs import __version__ as _version
from whylogs.api.writer import Writer
from whylogs.api.writer.writer import Writable
from whylogs.core import ColumnProfileView, DatasetProfileView
from whylogs.core.dataset_profile import DatasetProfile
from whylogs.core.errors import BadConfigError
from whylogs.core.metrics import Metric
from whylogs.core.metrics.compound_metric import CompoundMetric
from whylogs.core.utils import deprecated_alias
from whylogs.core.view.segmented_dataset_profile_view import SegmentedDatasetProfileView

try:
    from PIL.Image import Image as ImageType  # noqa trigger ImportError if unavailable

    from whylogs.extras.image_metric import ImageMetric
except ImportError:
    # dummy ImageMetric class so nothing will be an instance of it
    class ImageMetric:  # type: ignore[no-redef]
        pass


FIVE_MINUTES_IN_SECONDS = 60 * 5
logger = logging.getLogger(__name__)

API_KEY_ENV = "WHYLABS_API_KEY"


def _uncompound_metric_feature_flag() -> bool:
    return True


def _v0_compatible_image_feature_flag() -> bool:
    return False


def _uncompounded_column_name(column_name: str, metric_name: str, submetric_name: str, metric: Metric) -> str:
    if isinstance(metric, ImageMetric) and _v0_compatible_image_feature_flag():
        return submetric_name
    return f"{column_name}.{metric_name}.{submetric_name}"


def _uncompound_metric(col_name: str, metric_name: str, metric: CompoundMetric) -> Dict[str, ColumnProfileView]:
    result: Dict[str, ColumnProfileView] = dict()
    for submetric_name, submetric in metric.submetrics.items():
        new_col_name = _uncompounded_column_name(col_name, metric_name, submetric_name, metric)
        result[new_col_name] = ColumnProfileView({submetric.namespace: submetric})
    return result


def _uncompund_dataset_profile(prof: DatasetProfileView) -> DatasetProfileView:
    new_prof = DatasetProfileView(
        columns=prof._columns,
        dataset_timestamp=prof._dataset_timestamp,
        creation_timestamp=prof._creation_timestamp,
        metrics=prof._metrics,
    )
    new_columns: Dict[str, ColumnProfileView] = dict()
    for col_name, col_prof in new_prof._columns.items():
        for metric_name, metric in col_prof._metrics.items():
            if isinstance(metric, CompoundMetric):
                new_columns.update(_uncompound_metric(col_name, metric_name, metric))

    new_prof._columns.update(new_columns)
    return new_prof


class WhyLabsWriter(Writer):
    """
    A WhyLogs writer to upload DatasetProfileView's onto the WhyLabs platform.

    >**IMPORTANT**: In order to correctly send your profiles over, make sure you have
    the following environment variables set: `[WHYLABS_ORG_ID, WHYLABS_API_KEY, WHYLABS_DEFAULT_DATASET_ID]`. You
    can also set them with the option method or within the constructor, although it
    is highly recommended you don't persist credentials in code!

    Parameters
    ----------
    org_id: str, optional
        Your organization ID on the WhyLabs platform. If not informed, will get the mentioned environment
        variable instead.

    api_key: str, optional
        The API Key to authenticate with WhyLabs. If not informed, will get the mentioned environment
        variable instead.

    dataset_id: str, optional
        The Dataset ID, often referred to as model ID, that will tie the profile to your tracked data.
        If not informed, will get the mentioned environment variable instead.
    Returns
    -------
        None

    Examples
    --------
    An example usage of this method can be represented with the simple code above. Here we
    are assuming the user has already assigned a pandas.DataFrame to the `df` variable.

    ```python
    import whylogs as why

    profile = why.log(pandas=df)
    profile.writer("whylabs").write()
    ```

    """

    def __init__(self, org_id: Optional[str] = None, api_key: Optional[str] = None, dataset_id: Optional[str] = None):
        self._org_id = org_id or os.environ.get("WHYLABS_DEFAULT_ORG_ID")
        self._api_key = api_key or os.environ.get("WHYLABS_API_KEY")
        self._api_key_id = ""
        self._dataset_id = dataset_id or os.environ.get("WHYLABS_DEFAULT_DATASET_ID")
        self.whylabs_api_endpoint = os.environ.get("WHYLABS_API_ENDPOINT") or "https://api.whylabsapp.com"
        self._reference_profile_name = os.environ.get("WHYLABS_REFERENCE_PROFILE_NAME")
        self._api_log_client: Optional[ApiClient] = None
        self._ssl_ca_cert: Optional[str] = None
        self._configuration: Optional[Configuration] = None

    def check_interval(self, interval_seconds: int) -> None:
        if interval_seconds < FIVE_MINUTES_IN_SECONDS:
            raise BadConfigError("Bad WhyLabsWriter config: interval must be greater or equal to five minutes")

    def option(  # type: ignore
        self,
        org_id: Optional[str] = None,
        dataset_id: Optional[str] = None,
        api_key: Optional[str] = None,
        reference_profile_name: Optional[str] = None,
        configuration: Optional[Configuration] = None,
        ssl_ca_cert: Optional[str] = None,
    ) -> "WhyLabsWriter":
        """

        Parameters
        ----------
        org_id the organization ID
        dataset_id the dataset Id
        api_key the API key
        reference_profile_name the name of the reference profile
        configuration the additional configuration for the REST client

        Returns a "WhyLabsWriter" with these options configured
        -------

        """
        if dataset_id is not None:
            self._dataset_id = dataset_id
        if org_id is not None:
            self._org_id = org_id
        if api_key is not None:
            self._api_key = api_key
        if reference_profile_name is not None:
            self._reference_profile_name = reference_profile_name
        if configuration is not None:
            self._configuration = configuration
        if ssl_ca_cert is not None:
            self._ssl_ca_cert = ssl_ca_cert
        return self

    @deprecated_alias(profile="file")
    def write(self, file: Writable, **kwargs: Any) -> None:
        view = file.view() if isinstance(file, DatasetProfile) else file
        has_segments = isinstance(view, SegmentedDatasetProfileView)

        if not has_segments and not isinstance(view, DatasetProfileView):
            raise ValueError(
                "You must pass either a DatasetProfile or a DatasetProfileView in order to use this writer!"
            )

        if _uncompound_metric_feature_flag():
            if has_segments:
                updated_profile_view = _uncompund_dataset_profile(view.profile_view)
                view = SegmentedDatasetProfileView(
                    profile_view=updated_profile_view, segment=view._segment, partition=view._partition
                )

            else:
                view = _uncompund_dataset_profile(view)

        if kwargs.get("dataset_id") is not None:
            self._dataset_id = kwargs.get("dataset_id")

        with tempfile.NamedTemporaryFile() as tmp_file:
            if has_segments:
                view.write(path=tmp_file.name, use_v0=True)
            else:
                view.write(path=tmp_file.name)
            tmp_file.flush()

            dataset_timestamp = view.dataset_timestamp or datetime.datetime.now(datetime.timezone.utc)
            dataset_timestamp_epoch = int(dataset_timestamp.timestamp() * 1000)
            self._do_upload(
                dataset_timestamp=dataset_timestamp_epoch,
                profile_path=tmp_file.name,
            )

    def _validate_api_key(self) -> None:
        if self._api_key is None:
            raise ValueError("Missing API key. Set it via WHYLABS_API_KEY environment variable or as an api_key option")
        if len(self._api_key) < 12:
            raise ValueError("API key too short")
        if len(self._api_key) > 64:
            raise ValueError("API key too long")
        if self._api_key[10] != ".":
            raise ValueError("Invalid format. Expecting a dot at an index 10")
        self._api_key_id = self._api_key[:10]

    def _do_upload(
        self,
        dataset_timestamp: int,
        profile_path: str,
    ) -> requests.Response:
        if self._org_id is None:
            raise EnvironmentError(
                "Missing organization ID. Specify it via option or WHYLABS_DEFAULT_ORG_ID " "environment variable"
            )
        if self._dataset_id is None:
            raise EnvironmentError(
                "Missing dataset ID. Specify it via WHYLABS_DEFAULT_DATASET_ID environment "
                "variable or on your write method"
            )
        if self._api_key is None:
            raise EnvironmentError("Missing API key. Specify it via WHYLABS_API_KEY environment variable.")

        logger.debug("Generating the upload URL")
        upload_url = self._get_upload_url(dataset_timestamp=dataset_timestamp)
        api_key_id = self._api_key[:10] if self._api_key else None
        try:
            with open(profile_path, "rb") as f:
                http_response = requests.put(upload_url, data=f.read())
                if http_response.status_code == 200:
                    logger.info(
                        f"Done uploading {self._org_id}/{self._dataset_id}/{dataset_timestamp} to "
                        f"{self.whylabs_api_endpoint} with API token ID: {api_key_id}"
                    )
                return http_response
        except requests.RequestException as e:
            logger.info(
                f"Failed to upload {self._org_id}/{self._dataset_id}/{dataset_timestamp} to "
                + f"{self.whylabs_api_endpoint} with API token ID: {api_key_id}. Error occurred: {e}"
            )

    def _get_or_create_api_log_client(self) -> LogApi:
        environment_api_key = os.environ.get(API_KEY_ENV)

        if self._api_log_client is None:
            self._refresh_api_client()
        elif environment_api_key is not None and self._api_key != environment_api_key:
            logger.warning(f"Updating API key ID from: {self._api_key_id} to: {environment_api_key[:10]}")
            self._api_key = environment_api_key
            self._refresh_api_client()

        return LogApi(self._api_log_client)

    def _refresh_api_client(self) -> None:
        self._validate_api_key()

        if self._configuration is None:
            config = whylabs_client.Configuration()
        else:
            config = self._configuration
        config.ssl_ca_cert = self._ssl_ca_cert
        config.host = self.whylabs_api_endpoint
        config.api_key = {"ApiKeyAuth": self._api_key}
        config.discard_unknown_keys = True

        client = whylabs_client.ApiClient(config)
        client.user_agent = f"whylogs/python/{_version}"
        self._api_log_client = client

    def _get_rest_config(self) -> Configuration:
        if self._configuration is None:
            config = whylabs_client.Configuration()
        else:
            config = self._configuration

        config.host = self.whylabs_api_endpoint
        config.api_key = {"ApiKeyAuth": self._api_key}
        config.discard_unknown_keys = True
        return config

    @staticmethod
    def _build_log_async_request(dataset_timestamp: int) -> LogAsyncRequest:
        return LogAsyncRequest(dataset_timestamp=dataset_timestamp, segment_tags=[])

    @staticmethod
    def _build_log_reference_request(dataset_timestamp: int, alias: Optional[str] = None) -> LogReferenceRequest:
        return LogReferenceRequest(dataset_timestamp=dataset_timestamp, alias=alias)

    def _post_log_async(self, request: LogAsyncRequest, dataset_timestamp: int) -> AsyncLogResponse:
        log_api = self._get_or_create_api_log_client()
        try:
            result = log_api.log_async(org_id=self._org_id, dataset_id=self._dataset_id, log_async_request=request)
            return result
        except ForbiddenException as e:
            api_key_id = self._api_key[:10] if self._api_key else None
            logger.exception(
                f"Failed to upload {self._org_id}/{self._dataset_id}/{dataset_timestamp} to "
                f"{self.whylabs_api_endpoint}"
                f" with API token ID: {api_key_id}"
            )
            raise e

    def _post_log_reference(self, request: LogAsyncRequest, dataset_timestamp: int) -> LogReferenceResponse:
        log_api = self._get_or_create_api_log_client()
        try:
            async_result = log_api.log_reference(
                org_id=self._org_id, model_id=self._dataset_id, log_reference_request=request, async_req=True
            )
            result = async_result.get()
            return result
        except ForbiddenException as e:
            api_key_id = self._api_key[:10] if self._api_key else None
            logger.exception(
                f"Failed to upload {self._org_id}/{self._dataset_id}/{dataset_timestamp} to "
                f"{self.whylabs_api_endpoint}"
                f" with API token ID: {api_key_id}"
            )
            raise e

    def _get_upload_url(self, dataset_timestamp: int) -> str:
        if self._reference_profile_name is None:
            request = self._build_log_async_request(dataset_timestamp)
            res = self._post_log_async(request=request, dataset_timestamp=dataset_timestamp)
            upload_url = res["upload_url"]
        else:
            request = self._build_log_reference_request(dataset_timestamp, alias=self._reference_profile_name)
            res = self._post_log_reference(request=request, dataset_timestamp=dataset_timestamp)
            upload_url = res["upload_url"]
        return upload_url



import functools
import warnings
from typing import Any, Callable, Dict


def deprecated_alias(**aliases: str) -> Callable:
    def deprecated_decorator(method: Callable):
        @functools.wraps(method)
        def wrapper(*args, **kwargs):
            rename_kwargs(method.__name__, kwargs, aliases)
            return method(*args, **kwargs)

        return wrapper

    return deprecated_decorator


def rename_kwargs(func_name: str, kwargs: Dict[str, Any], aliases: Dict[str, str]):
    for alias, new in aliases.items():
        if alias in kwargs:
            if new in kwargs:
                raise TypeError(
                    f"{func_name} received both {alias} and {new} as arguments!"
                    f" {alias} is deprecated, use {new} instead."
                )
            warnings.warn(
                message=f"`{alias}` is deprecated as an argument to `{func_name}`; use" f" `{new}` instead.",
                category=DeprecationWarning,
                stacklevel=3,
            )
            kwargs[new] = kwargs.pop(alias)



import logging
import os
from typing import Any, Optional

from whylogs.api.writer import Writer
from whylogs.api.writer.writer import Writable
from whylogs.core.utils import deprecated_alias
from whylogs.core.view.segmented_dataset_profile_view import SegmentedDatasetProfileView

logger = logging.getLogger(__name__)


class LocalWriter(Writer):
    def __init__(self, base_dir: Optional[str] = None, base_name: Optional[str] = None) -> None:
        if base_dir is None:
            base_dir = os.getcwd()
        self._base_dir = base_dir
        self._base_name = base_name

    @deprecated_alias(profile="file")
    def write(
        self,
        file: Writable,
        dest: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        dest = dest or self._base_name or file.get_default_path()  # type: ignore
        full_path = os.path.join(self._base_dir, dest)
        has_segments = isinstance(file, SegmentedDatasetProfileView)
        if has_segments:
            file.write(full_path, use_v0=True)
        else:
            file.write(full_path)  # type: ignore

    def option(self, base_dir: Optional[str] = None, base_name: Optional[str] = None) -> "LocalWriter":  # type: ignore
        if base_dir is not None:
            self._base_dir = base_dir
        if base_name is not None:
            self._base_name = base_name
        return self



import logging
from typing import Any, Dict, List, Optional, TypeVar

from whylogs.core.configs import SummaryConfig
from whylogs.core.errors import DeserializationError, UnsupportedError
from whylogs.core.metrics import StandardMetric
from whylogs.core.metrics.metrics import _METRIC_DESERIALIZER_REGISTRY, Metric
from whylogs.core.proto import ColumnMessage, MetricComponentMessage, MetricMessage

logger = logging.getLogger(__name__)

METRIC = TypeVar("METRIC", bound=Metric)


class ColumnProfileView(object):
    def __init__(
        self,
        metrics: Dict[str, METRIC],
        success_count: int = 0,
        failure_count: int = 0,
    ):
        self._metrics: Dict[str, METRIC] = metrics.copy()
        self._success_count = success_count
        self._failure_count = failure_count

    def merge(self, other: "ColumnProfileView") -> "ColumnProfileView":
        all_names = set(self._metrics.keys()).union(other._metrics.keys())
        merged_metrics: Dict[str, Metric] = {}
        for n in all_names:
            lhs = self._metrics.get(n)
            rhs = other._metrics.get(n)

            res = lhs
            if lhs is None:
                res = rhs
            elif rhs is not None:
                res = lhs + rhs
            assert res is not None
            merged_metrics[n] = res
        return ColumnProfileView(
            metrics=merged_metrics,
            success_count=self._success_count + other._success_count,
            failure_count=self._failure_count + other._failure_count,
        )

    def __add__(self, other: "ColumnProfileView") -> "ColumnProfileView":
        return self.merge(other)

    def __getstate__(self) -> bytes:
        return self.serialize()

    def __setstate__(self, serialized_profile: bytes):
        profile = ColumnProfileView.deserialize(serialized_profile)
        self.__dict__.update(profile.__dict__)

    def serialize(self) -> bytes:
        return self.to_protobuf().SerializeToString()

    @classmethod
    def deserialize(cls, serialized_profile: bytes) -> "ColumnProfileView":
        column_message = ColumnMessage.FromString(serialized_profile)
        return ColumnProfileView.from_protobuf(column_message)

    def get_metric(self, m_name: str) -> Optional[METRIC]:
        return self._metrics.get(m_name)

    def to_protobuf(self) -> ColumnMessage:
        res: Dict[str, MetricComponentMessage] = {}
        for m_name, m in self._metrics.items():
            for mc_name, mc in m.to_protobuf().metric_components.items():
                res[f"{m.namespace}/{mc_name}"] = mc
        return ColumnMessage(metric_components=res)

    def get_metric_component_paths(self) -> List[str]:
        res: List[str] = []
        for m_name, m in self._metrics.items():
            for mc_name in m.get_component_paths():
                res.append(f"{m.namespace}/{mc_name}")
        return res

    def get_metric_names(self) -> List[str]:
        return [name for name in self._metrics]

    def to_summary_dict(
        self, *, column_metric: Optional[str] = None, cfg: Optional[SummaryConfig] = None
    ) -> Dict[str, Any]:
        if cfg is None:
            cfg = SummaryConfig()

        res = {}
        for metric_name, metric in self._metrics.items():
            if column_metric is not None and metric_name != column_metric:
                continue
            try:
                m_sum = metric.to_summary_dict(cfg=cfg)
                for k, v in m_sum.items():
                    res[f"{metric_name}/{k}"] = v
            except NotImplementedError:
                logger.warning(f"No summary implemented for {metric_name}")

        if column_metric is not None and len(res) == 0:
            raise UnsupportedError(f"No metric available for requested column metric: {column_metric}")

        return res

    @classmethod
    def zero(cls, msg: ColumnMessage) -> "ColumnProfileView":
        return ColumnProfileView(metrics={})

    @classmethod
    def from_protobuf(cls, msg: ColumnMessage) -> "ColumnProfileView":
        result_metrics: Dict[str, Metric] = {}
        metric_messages: Dict[str, Dict[str, MetricComponentMessage]] = {}
        for full_path, c_msg in msg.metric_components.items():
            metric_name = full_path.split("/")[0]
            metric_components = metric_messages.get(metric_name)

            if metric_components is None:
                metric_components = {}
            metric_messages[metric_name] = metric_components

            c_key = full_path[len(metric_name) + 1 :]
            metric_components[c_key] = c_msg

        for m_name, metric_components in metric_messages.items():
            m_enum = StandardMetric.__members__.get(m_name)
            if m_enum is None:
                if m_name in _METRIC_DESERIALIZER_REGISTRY:
                    metric_class = _METRIC_DESERIALIZER_REGISTRY[m_name]
                else:
                    raise UnsupportedError(f"Unsupported metric: {m_name}")
            else:
                metric_class = m_enum.value

            m_msg = MetricMessage(metric_components=metric_components)
            try:
                result_metrics[m_name] = metric_class.from_protobuf(m_msg)
            except:  # noqa
                raise DeserializationError(f"Failed to deserialize metric: {m_name}")
        return ColumnProfileView(metrics=result_metrics)

    @classmethod
    def from_bytes(cls, data: bytes) -> "ColumnProfileView":
        msg = ColumnMessage()
        if isinstance(data, bytearray):
            data = bytes(data)
        msg.ParseFromString(data)
        return ColumnProfileView.from_protobuf(msg)



import logging
import os.path
import time
from datetime import datetime, timezone
from typing import Any, Dict, Mapping, Optional, Union

from whylogs.api.writer.writer import Writable
from whylogs.core.metrics import Metric
from whylogs.core.model_performance_metrics.model_performance_metrics import (
    ModelPerformanceMetrics,
)
from whylogs.core.utils.utils import deprecated_alias

from .column_profile import ColumnProfile
from .input_resolver import _pandas_or_dict
from .schema import DatasetSchema
from .stubs import pd
from .view import DatasetProfileView

logger = logging.getLogger(__name__)

_LARGE_CACHE_SIZE_LIMIT = 1024 * 100
_MODEL_PERFORMANCE_KEY = "model_performance_metrics"


class DatasetProfile(Writable):
    """
    Dataset profile represents a collection of in-memory profiling stats for a dataset.

    Args:
        schema: :class:`DatasetSchema`, optional
            An object that represents the data column names and types
        dataset_timestamp: int, optional
            A timestamp integer that best represents the date tied to the dataset generation.
            i.e.: A January 1st 2019 Sales Dataset will have 1546300800000 as the timestamp in miliseconds (UTC).
            If None is provided, it will take the current timestamp as default
        creation_timestamp: int, optional
            The timestamp tied to the exact moment when the :class:`DatasetProfile` is created.
            If None is provided, it will take the current timestamp as default
    """

    def __init__(
        self,
        schema: Optional[DatasetSchema] = None,
        dataset_timestamp: Optional[datetime] = None,
        creation_timestamp: Optional[datetime] = None,
        metrics: Optional[Dict[str, Union[Metric, Any]]] = None,
    ):
        if schema is None:
            schema = DatasetSchema()
        now = datetime.utcnow()
        self._dataset_timestamp = dataset_timestamp or now
        self._creation_timestamp = creation_timestamp or now
        self._schema = schema
        self._columns: Dict[str, ColumnProfile] = dict()
        self._is_active = False
        self._track_count = 0
        new_cols = schema.get_col_names()
        self._initialize_new_columns(new_cols)
        self._metrics: Dict[str, Union[Metric, Any]] = metrics or dict()

    @property
    def creation_timestamp(self) -> datetime:
        return self._creation_timestamp

    @property
    def dataset_timestamp(self) -> datetime:
        return self._dataset_timestamp

    @property
    def is_active(self) -> bool:
        """Returns True if the profile tracking code is currently running."""
        return self._is_active

    @property
    def is_empty(self) -> bool:
        """Returns True if the profile tracking code is currently running."""
        return self._track_count == 0

    def set_dataset_timestamp(self, dataset_timestamp: datetime) -> None:
        if dataset_timestamp.tzinfo is None:
            logger.warning("No timezone set in the datetime_timestamp object. Default to local timezone")

        self._dataset_timestamp = dataset_timestamp.astimezone(tz=timezone.utc)

    def add_metric(self, col_name: str, metric: Metric) -> None:
        if col_name not in self._columns:
            raise ValueError(f"{col_name} is not a column in the dataset profile")
        self._columns[col_name].add_metric(metric)

    def add_dataset_metric(self, name: str, metric: Metric) -> None:
        self._metrics[name] = metric

    def add_model_performance_metrics(self, metric: ModelPerformanceMetrics) -> None:
        self._metrics[_MODEL_PERFORMANCE_KEY] = metric

    @property
    def model_performance_metrics(self) -> ModelPerformanceMetrics:
        if self._metrics:
            return self._metrics.get(_MODEL_PERFORMANCE_KEY)
        return None

    def track(
        self,
        obj: Any = None,
        *,
        pandas: Optional[pd.DataFrame] = None,
        row: Optional[Mapping[str, Any]] = None,
    ) -> None:
        try:
            self._is_active = True
            self._track_count += 1
            self._do_track(obj, pandas=pandas, row=row)
        finally:
            self._is_active = False

    def _do_track(
        self,
        obj: Any = None,
        *,
        pandas: Optional[pd.DataFrame] = None,
        row: Optional[Mapping[str, Any]] = None,
    ) -> None:

        pandas, row = _pandas_or_dict(obj, pandas, row)

        # TODO: do this less frequently when operating at row level
        dirty = self._schema.resolve(pandas=pandas, row=row)
        if dirty:
            schema_col_keys = self._schema.get_col_names()
            new_cols = (col for col in schema_col_keys if col not in self._columns)
            self._initialize_new_columns(tuple(new_cols))

        if pandas is not None:
            # TODO: iterating over each column in order assumes single column metrics
            #   but if we instead iterate over a new artifact contained in dataset profile: "MetricProfiles", then
            #   each metric profile can specify which columns its tracks, and we can call like this:
            #   metric_profile.track(pandas)
            for k in pandas.keys():
                self._columns[k].track_column(pandas[k])
            return

        if row is not None:
            for k in row.keys():
                self._columns[k].track_column([row[k]])
            return

        raise NotImplementedError

    def _track_classification_metrics(
        self,
        targets,
        predictions,
        scores=None,
        target_field=None,
        prediction_field=None,
        score_field=None,
    ) -> None:
        """
        Function to track metrics based on validation data.
        user may also pass the associated attribute names associated with
        target, prediction, and/or score.
        Parameters
        ----------
        targets : List[Union[str, bool, float, int]]
            actual validated values
        predictions : List[Union[str, bool, float, int]]
            inferred/predicted values
        scores : List[float], optional
            assocaited scores for each inferred, all values set to 1 if not
            passed
        target_field : str, optional
            Description
        prediction_field : str, optional
            Description
        score_field : str, optional
            Description
        target_field : str, optional
        prediction_field : str, optional
        score_field : str, optional
        """
        if not self.model_performance_metrics:
            self.add_model_performance_metrics(ModelPerformanceMetrics())
        self.model_performance_metrics.compute_confusion_matrix(
            predictions=predictions,
            targets=targets,
            scores=scores,
            target_field=target_field,
            prediction_field=prediction_field,
            score_field=score_field,
        )

    def _track_regression_metrics(
        self,
        targets,
        predictions,
        scores=None,
        target_field=None,
        prediction_field=None,
        score_field=None,
    ) -> None:
        """
        Function to track regression metrics based on validation data.
        user may also pass the associated attribute names associated with
        target, prediction, and/or score.
        Parameters
        ----------
        targets : List[Union[str, bool, float, int]]
            actual validated values
        predictions : List[Union[str, bool, float, int]]
            inferred/predicted values
        scores : List[float], optional
            assocaited scores for each inferred, all values set to 1 if not
            passed
        target_field : str, optional
            Description
        prediction_field : str, optional
            Description
        score_field : str, optional
            Description
        target_field : str, optional
        prediction_field : str, optional
        score_field : str, optional
        """
        if not self.model_performance_metrics:
            self.add_model_performance_metrics(ModelPerformanceMetrics())
        self.model_performance_metrics.compute_regression_metrics(
            predictions=predictions,
            targets=targets,
            scores=scores,
            target_field=target_field,
            prediction_field=prediction_field,
            score_field=score_field,
        )

    def _initialize_new_columns(self, new_cols: tuple) -> None:
        for col in new_cols:
            col_schema = self._schema.get(col)
            if col_schema:
                self._columns[col] = ColumnProfile(name=col, schema=col_schema, cache_size=self._schema.cache_size)
            else:
                logger.warning("Encountered a column without schema: %s", col)

    def view(self) -> DatasetProfileView:
        columns = {}
        for c_name, c in self._columns.items():
            columns[c_name] = c.view()
        return DatasetProfileView(
            columns=columns,
            dataset_timestamp=self.dataset_timestamp,
            creation_timestamp=self.creation_timestamp,
            metrics=self._metrics,
        )

    def flush(self) -> None:
        for col in self._columns.values():
            col.flush()

    @staticmethod
    def get_default_path(path) -> str:
        if not path.endswith("bin"):
            path = os.path.join(path, f"profile.{int(round(time.time() * 1000))}.bin")
        return path

    @deprecated_alias(path_or_base_dir="path")
    def write(self, path: Optional[str] = None, **kwargs: Any) -> None:
        output_path = self.get_default_path(path=path)
        self.view().write(output_path)
        logger.debug("Wrote profile to path: %s", output_path)

    @classmethod
    def read(cls, input_path: str) -> DatasetProfileView:
        return DatasetProfileView.read(input_path)

    def __repr__(self) -> str:
        return f"DatasetProfile({len(self._columns)} columns). Schema: {str(self._schema)}"



from logging import getLogger
from typing import Dict, List, Optional, Set, Union

from whylogs.core.metrics.metrics import Metric
from whylogs.core.proto.v0 import ModelMetricsMessage, ModelProfileMessage, ModelType

from .confusion_matrix import ConfusionMatrix
from .regression_metrics import RegressionMetrics

OUTPUT_FIELD_CATEGORY = "output"
logger = getLogger(__name__)


class ModelPerformanceMetrics:
    """
    Container class for various model-related performance metrics

    Attributes:
        confusion_matrix (ConfusionMatrix): ConfusionMatrix which keeps it track of counts with NumberTracker
        regression_metrics (RegressionMetrics): Regression Metrics keeps track of a common regression metrics in case the targets are continous.
    """

    def __init__(
        self,
        confusion_matrix: Optional[ConfusionMatrix] = None,
        regression_metrics: Optional[RegressionMetrics] = None,
        metrics: Optional[Dict[str, Metric]] = None,
        field_metadata: Optional[Dict[str, Set[str]]] = None,
    ):
        self.model_type = None
        self.confusion_matrix = confusion_matrix
        self.regression_metrics = regression_metrics
        self.metrics = metrics
        self.field_metadata = field_metadata

    def to_protobuf(
        self,
    ) -> ModelProfileMessage:

        model_type = ModelType.UNKNOWN
        if not self.model_type:
            if self.confusion_matrix:
                model_type = ModelType.CLASSIFICATION
            elif self.regression_metrics:
                model_type = ModelType.REGRESSION
        else:
            model_type = self.model_type

        model_metrics = ModelMetricsMessage(
            scoreMatrix=self.confusion_matrix.to_protobuf() if self.confusion_matrix else None,
            regressionMetrics=self.regression_metrics.to_protobuf() if self.regression_metrics else None,
            modelType=model_type,
        )

        return ModelProfileMessage(output_fields=self.output_fields, metrics=model_metrics)

    @classmethod
    def from_protobuf(
        cls,
        message: ModelProfileMessage,
    ) -> "ModelPerformanceMetrics":
        # TODO: update format to support storing other field metadata, for now just support output field in v0 message format
        output_fields = (
            {column_name: set([OUTPUT_FIELD_CATEGORY]) for column_name in message.output_fields}
            if message.output_fields
            else None
        )
        confusion_matrix = None
        regression_metrics = None
        if message.metrics is None:
            logger.warning("deserializing a ModelPerformanceMetrics without a metrics field")
        else:
            confusion_matrix = ConfusionMatrix.from_protobuf(message.metrics.scoreMatrix)
            regression_metrics = RegressionMetrics.from_protobuf(message.metrics.regressionMetrics)

        return ModelPerformanceMetrics(
            confusion_matrix=confusion_matrix, regression_metrics=regression_metrics, field_metadata=output_fields
        )

    def compute_confusion_matrix(
        self,
        predictions: List[Union[str, int, bool, float]],
        targets: List[Union[str, int, bool, float]],
        scores: List[float] = None,
    ):
        """
        computes the confusion_matrix, if one is already present merges to old one.

        Args:
            predictions (List[Union[str, int, bool]]):
            targets (List[Union[str, int, bool]]):
            scores (List[float], optional):
        """
        labels = sorted(list(set(targets + predictions)))
        confusion_matrix = ConfusionMatrix(labels=labels)
        confusion_matrix.add(predictions, targets, scores)

        if self.confusion_matrix is None or self.confusion_matrix.labels is None or self.confusion_matrix.labels == []:
            self.confusion_matrix = confusion_matrix
        else:
            self.confusion_matrix = self.confusion_matrix.merge(confusion_matrix)

    def compute_regression_metrics(
        self,
        predictions: List[Union[float, int]],
        targets: List[Union[float, int]],
    ):
        regression_metrics = RegressionMetrics()
        regression_metrics.add(predictions, targets)
        if self.regression_metrics:
            self.regression_metrics = self.regression_metrics.merge(regression_metrics)
        else:
            self.regression_metrics = regression_metrics

    def add_metadata_to_field(self, column_name: str, categories: Set[str]) -> None:
        if not self.field_metadata:
            self.field_metadata = dict()
        self.field_metadata[column_name] = categories

    def _set_column_value(self, column_name, value):
        if column_name not in self.field_metadata:
            self.field_metadata[column_name] = set([value])
        else:
            self.field_metadata[column_name].add(value)

    def specify_output_fields(self, column_names: Union[str, Set[str]]) -> None:
        if not self.field_metadata:
            self.field_metadata = dict()
        if isinstance(column_names, str):
            self._set_column_value(column_names, OUTPUT_FIELD_CATEGORY)
        else:
            for column_name in column_names:
                self._set_column_value(column_name, OUTPUT_FIELD_CATEGORY)

    @property
    def output_fields(self) -> Optional[List[str]]:
        output_column_names = None
        if self.field_metadata:
            for column_name in self.field_metadata:
                field_categories = self.field_metadata[column_name]
                if OUTPUT_FIELD_CATEGORY in field_categories:
                    if output_column_names is None:
                        output_column_names = []
                    output_column_names.append(column_name)
        return output_column_names

    def merge(self, other):
        """
        :type other: ModelMetrics
        """
        if other is None or (other.confusion_matrix is None and other.regression_metrics is None):
            return self

        merged_field_metadata: Optional[Dict[str, Set[str]]] = None
        if self.field_metadata is not None:
            if other.field_metadata is None:
                merged_field_metadata = self.field_metadata
            else:
                merged_keys = set(self.field_metadata.keys() + other.field_metadata.keys)
                merged_field_metadata = dict()
                for key in merged_keys:
                    categories = self.field_metadata.get(key)
                    other_categories = other.field_metadata.get(key)
                    if categories and other_categories:
                        merged_field_metadata[key] = categories.union(other_categories)
                    elif categories:
                        merged_field_metadata[key] = categories
                    else:
                        merged_field_metadata[key] = other_categories

        else:
            merged_field_metadata = other.field_metadata
        self.field_metadata = merged_field_metadata
        return ModelPerformanceMetrics(
            confusion_matrix=self.confusion_matrix.merge(other.confusion_matrix)
            if self.confusion_matrix
            else other.confusion_matrix,
            regression_metrics=self.regression_metrics.merge(other.regression_metrics)
            if self.regression_metrics
            else other.regression_metrics,
            field_metadata=merged_field_metadata,
        )



from typing import Optional

from whylogs.api.logger.result_set import ResultSet
from whylogs.api.reader.reader import Reader


class LocalReader(Reader):
    def __init__(self, path: Optional[str] = None):
        self.path = path

    def read(self, path: Optional[str] = None, **kwargs) -> ResultSet:
        file_path = path or self.path
        if file_path is None:
            raise ValueError("You must define a path to read your file from!")
        return self.get_file_from_path(path=file_path)

    def option(self, path: Optional[str]) -> "LocalReader":
        if path is not None:
            self.path = path
        return self



import datetime
from logging import getLogger
from typing import Dict, List, Optional, Tuple

import whylogs_sketching as ds  # type: ignore

from whylogs.core import ColumnProfileView, DatasetProfileView
from whylogs.core.errors import DeserializationError
from whylogs.core.metrics import (
    ColumnCountsMetric,
    DistributionMetric,
    FrequentItemsMetric,
    IntsMetric,
    StandardMetric,
    TypeCountersMetric,
)
from whylogs.core.metrics.metric_components import (
    FractionalComponent,
    FrequentStringsComponent,
    IntegralComponent,
    KllComponent,
    MaxIntegralComponent,
    MinIntegralComponent,
)
from whylogs.core.metrics.metrics import CardinalityMetric
from whylogs.core.proto import SegmentTag
from whylogs.core.proto.v0 import (
    ColumnMessageV0,
    CountersV0,
    DatasetProfileMessageV0,
    DatasetPropertiesV0,
    FrequentItemsSketchMessageV0,
    HllSketchMessageV0,
    InferredType,
    NumbersMessageV0,
    SchemaMessageV0,
    VarianceMessage,
)
from whylogs.core.segment import Segment
from whylogs.core.segmentation_partition import SegmentationPartition
from whylogs.core.utils import read_delimited_protobuf
from whylogs.core.utils.timestamp_calculations import to_utc_milliseconds

_DEFAULT_V0_LG_MAX_K = 12
_DEFAULT_V0_KLL_K = 128
_DEFAULT_V0_FI_LG_K = 10

EMPTY_FI_SKETCH: bytes = ds.frequent_strings_sketch(_DEFAULT_V0_FI_LG_K).serialize()
EMPTY_HLL: bytes = ds.hll_sketch(_DEFAULT_V0_LG_MAX_K).serialize_compact()
EMPTY_KLL: bytes = ds.kll_doubles_sketch(k=_DEFAULT_V0_KLL_K).serialize()
_empty_theta_union = ds.theta_union()
_empty_theta_union.update(ds.update_theta_sketch())
EMPTY_THETA: bytes = _empty_theta_union.get_result().serialize()

logger = getLogger(__name__)

PARTITION_ID = "segp_id"
PARTITION_NAME = "segp_name"
PARTITION_HAS_FILTER = "segp_filter"
SEGMENT_ON_COLUMN = "segp_col"
SEGMENT_ON_COLUMNS = "segp_cols"
_TAG_PREFIX = "whylogs.tag."


def _generate_segment_tags_metadata(
    segment: Segment, partition: SegmentationPartition
) -> Tuple[Dict[str, str], List[SegmentTag], Dict[str, str]]:
    segment_metadata: Optional[Dict[str, str]] = None
    segment_tags: Optional[List[SegmentTag]] = None
    if partition.mapper and partition.mapper.col_names:
        segment_metadata = {}
        segment_metadata[PARTITION_ID] = partition.id
        segment_metadata[PARTITION_NAME] = partition.name
        if partition.filter:
            segment_metadata[PARTITION_HAS_FILTER] = partition.filter_id

        segment_tags = []
        col_names = partition.mapper.col_names
        index = 0
        for column_name in col_names:
            segment_tags.append(SegmentTag(key=_TAG_PREFIX + column_name, value=segment.key[index]))
    else:
        raise NotImplementedError(
            "serialization of custom v1 segmented profiles not yet supported! please use columnar segmentation. "
            f"Found partition: {partition} and segments: {segment}"
        )

    segment_message_tags: Dict[str, str] = {}
    for segment_tag in segment_tags:
        segment_message_tags[segment_tag.key] = segment_tag.value

    if partition.mapper.map:
        raise NotImplementedError(
            "custom mapping function segment serialization not yet supported, please use column based segmentation: "
            f"for partition: {partition} and segments: {segment}"
        )

    if partition.mapper.col_names:
        segment_columns = partition.mapper.col_names
        if len(segment_columns) == 1:
            segment_metadata[SEGMENT_ON_COLUMN] = segment_columns[0]
        elif len(segment_columns) > 1:
            segment_metadata[SEGMENT_ON_COLUMNS] = "".join(["(" + str(col) + ")" for col in segment_columns])

    return segment_message_tags, segment_tags, segment_metadata


def read_v0_to_view(path: str) -> DatasetProfileView:
    with open(path, "r+b") as f:
        v0_msg = read_delimited_protobuf(f, DatasetProfileMessageV0)
        if v0_msg is None:
            raise DeserializationError("Unexpected empty message")
        return v0_to_v1_view(v0_msg)


def v0_to_v1_view(msg: DatasetProfileMessageV0) -> DatasetProfileView:
    columns: Dict[str, ColumnProfileView] = {}

    dataset_timestamp = datetime.datetime.fromtimestamp(msg.properties.data_timestamp / 1000.0)
    creation_timestamp = datetime.datetime.fromtimestamp(msg.properties.session_timestamp / 1000.0)

    for col_name, col_msg in msg.columns.items():
        dist_metric = _extract_dist_metric(col_msg)
        fs = FrequentStringsComponent(ds.frequent_strings_sketch.deserialize(col_msg.frequent_items.sketch))
        fi_metric = FrequentItemsMetric(frequent_strings=fs)
        count_metrics = _extract_col_counts(col_msg)
        type_counters_metric = _extract_type_counts_metric(col_msg)
        int_metric = _extract_ints_metric(col_msg)

        columns[col_name] = ColumnProfileView(
            metrics={
                StandardMetric.distribution.name: dist_metric,
                StandardMetric.frequent_items.name: fi_metric,
                StandardMetric.counts.name: count_metrics,
                StandardMetric.types.name: type_counters_metric,
                StandardMetric.cardinality.name: type_counters_metric,
                StandardMetric.ints.name: int_metric,
            }
        )

    return DatasetProfileView(
        columns=columns, dataset_timestamp=dataset_timestamp, creation_timestamp=creation_timestamp
    )


def _extract_ints_metric(msg: ColumnMessageV0) -> IntsMetric:
    int_max = msg.numbers.longs.max
    int_min = msg.numbers.longs.min
    return IntsMetric(max=MaxIntegralComponent(int_max), min=MinIntegralComponent(int_min))


def _extract_numbers_message_v0(col_prof: ColumnProfileView) -> NumbersMessageV0:
    distribution_metric: DistributionMetric = col_prof.get_metric(DistributionMetric.get_namespace())
    variance_message = VarianceMessage()
    msg_v0 = NumbersMessageV0(histogram=EMPTY_KLL, compact_theta=EMPTY_THETA, variance=variance_message)

    if distribution_metric is None:
        return msg_v0

    counts_metric: ColumnCountsMetric = col_prof.get_metric(ColumnCountsMetric.get_namespace())
    variance_message.count = counts_metric.n.value

    kll = distribution_metric.kll.value
    if not kll.is_empty():
        variance_message.mean = distribution_metric.mean.value
        variance_message.sum = distribution_metric.m2.value
        msg_v0 = NumbersMessageV0(histogram=kll.serialize(), compact_theta=EMPTY_THETA, variance=variance_message)

    return msg_v0


def _extract_type_counts_metric(msg: ColumnMessageV0) -> TypeCountersMetric:
    int_count = msg.schema.typeCounts.get(InferredType.INTEGRAL)
    bool_count = msg.schema.typeCounts.get(InferredType.BOOLEAN)
    frac_count = msg.schema.typeCounts.get(InferredType.FRACTIONAL)
    string_count = msg.schema.typeCounts.get(InferredType.STRING)
    obj_count = msg.schema.typeCounts.get(InferredType.UNKNOWN)
    return TypeCountersMetric(
        integral=IntegralComponent(int_count or 0),
        fractional=IntegralComponent(frac_count or 0),
        boolean=IntegralComponent(bool_count or 0),
        string=IntegralComponent(string_count or 0),
        object=IntegralComponent(obj_count or 0),
    )


def _extract_schema_message_v0(col_prof: ColumnProfileView) -> SchemaMessageV0:
    metric: TypeCountersMetric = col_prof.get_metric(TypeCountersMetric.get_namespace())
    if metric is None:
        return SchemaMessageV0()

    type_counts: Dict[int, int] = {}
    type_counts[InferredType.INTEGRAL] = metric.integral.value
    type_counts[InferredType.BOOLEAN] = metric.boolean.value
    type_counts[InferredType.FRACTIONAL] = metric.fractional.value
    type_counts[InferredType.STRING] = metric.string.value
    type_counts[InferredType.UNKNOWN] = metric.object.value

    msg_v0 = SchemaMessageV0(
        typeCounts=type_counts,
    )

    return msg_v0


def _extract_col_counts(msg: ColumnMessageV0) -> ColumnCountsMetric:
    count_n = msg.counters.count
    count_null = msg.counters.null_count
    return ColumnCountsMetric(n=IntegralComponent(count_n or 0), null=IntegralComponent(count_null.value or 0))


def _extract_counters_v0(col_prof: ColumnProfileView) -> CountersV0:
    counts_metric: ColumnCountsMetric = col_prof.get_metric(ColumnCountsMetric.get_namespace())
    count_options = dict(count=0)

    if counts_metric is None:
        return CountersV0(**dict(count=0))

    count_options = dict(count=counts_metric.n.value)

    return CountersV0(**count_options)


def _extract_dist_metric(msg: ColumnMessageV0) -> DistributionMetric:
    kll_bytes = msg.numbers.histogram
    floats_sk = None
    doubles_sk: Optional[ds.kll_doubles_sketch] = None
    try:
        floats_sk = ds.kll_floats_sketch.deserialize(kll_bytes)
    except ValueError as e:
        logger.info(f"kll encountered old format which threw exception: {e}, attempting kll_doubles deserialization.")
    if floats_sk is None:
        doubles_sk = ds.kll_doubles_sketch.deserialize(kll_bytes)
    else:
        doubles_sk = ds.kll_floats_sketch.float_to_doubles(floats_sk)

    dist_mean = msg.numbers.variance.mean
    dist_m2 = msg.numbers.variance.sum
    return DistributionMetric(
        kll=KllComponent(doubles_sk),
        mean=FractionalComponent(dist_mean),
        m2=FractionalComponent(dist_m2),
    )


def _extract_frequent_items_sketch_message_v0(col_prof: ColumnProfileView) -> FrequentItemsSketchMessageV0:
    frequent_items_metric: FrequentItemsMetric = col_prof.get_metric(FrequentItemsMetric.get_namespace())
    frequent_items_v0 = FrequentItemsSketchMessageV0(sketch=EMPTY_FI_SKETCH)
    if frequent_items_metric is None:
        return frequent_items_v0

    frequent_items_v0.sketch = frequent_items_metric.frequent_strings.value.serialize()
    frequent_items_v0.lg_max_k = _DEFAULT_V0_LG_MAX_K

    return frequent_items_v0


def _extract_hll_sketch_message_v0(col_prof) -> HllSketchMessageV0:
    cardinality_metric: CardinalityMetric = col_prof.get_metric(CardinalityMetric.get_namespace())
    hll_v0 = HllSketchMessageV0(sketch=EMPTY_HLL)
    if cardinality_metric is None:
        return hll_v0
    hll_sketch = cardinality_metric.hll.value
    if not hll_sketch.is_empty():
        hll_v0.sketch = hll_sketch.serialize_compact()
        hll_v0.lg_k = _DEFAULT_V0_LG_MAX_K

    return hll_v0


def v1_to_dataset_profile_message_v0(
    profile_view: DatasetProfileView,
    segment: Optional[Segment],
    partition: Optional[SegmentationPartition],
) -> DatasetProfileMessageV0:

    segment_message_tags: Optional[Dict[str, str]] = None
    segment_tags: Optional[List[SegmentTag]] = None
    segment_metadata: Optional[Dict[str, str]] = None

    if partition and partition.mapper and partition.mapper.col_names:
        segment_message_tags, segment_tags, segment_metadata = _generate_segment_tags_metadata(segment, partition)
    elif partition:
        raise NotImplementedError(
            f"Conversion of custom v1 segmented profiles to v0 not supported! please use column value segmentation, found tags: {segment_tags}"
        )

    properties_v0 = DatasetPropertiesV0(
        schema_major_version=1,  # https://github.com/whylabs/whylogs/blob/maintenance/0.7.x/src/whylogs/core/datasetprofile.py#L37-L38
        schema_minor_version=2,
        data_timestamp=to_utc_milliseconds(profile_view._dataset_timestamp),
        session_timestamp=to_utc_milliseconds(profile_view._creation_timestamp),
        metadata=segment_metadata,
        tags=segment_message_tags,
    )

    columns_v0: Dict[str, ColumnMessageV0] = {}
    columns = profile_view.get_columns()
    for col_name, col_prof in columns.items():
        columns_v0[col_name] = ColumnMessageV0(
            name=col_name,
            counters=_extract_counters_v0(col_prof),
            schema=_extract_schema_message_v0(col_prof),
            numbers=_extract_numbers_message_v0(col_prof),
            frequent_items=_extract_frequent_items_sketch_message_v0(col_prof),
            cardinality_tracker=_extract_hll_sketch_message_v0(col_prof),
        )
    metrics = profile_view.model_performance_metrics
    model_profile_message = metrics.to_protobuf() if metrics else None

    message = DatasetProfileMessageV0(
        properties=properties_v0,
        columns=columns_v0,
        modeProfile=model_profile_message,
    )
    return message



from logging import getLogger
from typing import List, Optional

from typing_extensions import TypedDict
from whylogs_sketching import kll_doubles_sketch  # type: ignore

from whylogs.core.metrics import DistributionMetric
from whylogs.core.view.column_profile_view import ColumnProfileView
from whylogs.viz.utils import _calculate_bins

logger = getLogger(__name__)

MAX_HIST_BUCKETS = 30
HIST_AVG_NUMBER_PER_BUCKET = 4.0


class HistogramSummary(TypedDict):
    start: float
    end: float
    width: float
    counts: List[int]
    max: float
    min: float
    bins: List[float]
    n: int


def histogram_from_view(column_view: ColumnProfileView, feature_name: str) -> HistogramSummary:
    col_dist: Optional[DistributionMetric] = column_view.get_metric("distribution")
    if not col_dist:
        raise ValueError("Distribution Metrics not found for feature {}.".format(feature_name))

    target_kill = col_dist.kll.value
    target_histogram = _histogram_from_sketch(target_kill)
    return target_histogram


def _histogram_from_sketch(
    sketch: kll_doubles_sketch, max_buckets: int = None, avg_per_bucket: Optional[float] = None
) -> HistogramSummary:
    """
    Generate a summary of a kll_floats_sketch, including a histogram

    Parameters
    ----------
    sketch : kll_floats_sketch
        Data sketch
    max_buckets : int
        Override the default maximum number of buckets
    avg_per_bucket : int
        Override the default target number of items per bucket.

    Returns
    -------
    histogram : HistogramSummary
        Protobuf histogram message
    """
    n = sketch.get_n()
    start = sketch.get_min_value()
    max_val = sketch.get_max_value()
    end = max_val
    if max_buckets is None:
        max_buckets = MAX_HIST_BUCKETS
    if avg_per_bucket is None:
        avg_per_bucket = HIST_AVG_NUMBER_PER_BUCKET

    if (n < 2) or (start == end):
        dx = abs(start) * 1e-7
        end = start + dx
        bins = [start, end]
        counts = [n]
    else:
        bins, end = _calculate_bins(end, start, n, avg_per_bucket, max_buckets)
        pmf = sketch.get_pmf(bins)
        counts = [round(p * n) for p in pmf]
        counts = counts[1:-1]

    histogram: HistogramSummary = {
        "start": start,
        "end": end,
        "width": 0,
        "counts": counts,
        "max": max_val,
        "min": start,
        "bins": bins,
        "n": n,
    }
    return histogram



from whylogs.core.configs import SummaryConfig

from ..metric_constraints import MetricConstraint, MetricsSelector


def frequent_strings_in_reference_set(column_name: str, reference_set: dict) -> MetricConstraint:
    """Determine whether a set of variables appear in the frequent strings for a string column.
    Every item in frequent strings must be in defined reference set

    Parameters
    ----------
    column_name : str
        Columns the constraint is applied to.
    reference_set : dict
        Reference set for applying the constraint
    """
    frequent_strings = MetricsSelector(metric_name="frequent_items", column_name=column_name)

    def labels_in_set(metric):
        frequent_strings = metric.to_summary_dict(SummaryConfig())["frequent_strings"]
        result = all(item.value in reference_set for item in frequent_strings)
        return result

    constraint_name = f"{column_name} values in set {reference_set}"
    constraint = MetricConstraint(name=constraint_name, condition=labels_in_set, metric_selector=frequent_strings)
    return constraint


def n_most_common_items_in_set(column_name: str, n: int, reference_set: dict) -> MetricConstraint:
    """Validate if the top n most common items appear in the dataset.

    Parameters
    ----------
    column_name : str
        Columns the constraint is applied to.
    n : int
        n most common items or strings.
    reference_set : dict
        Reference set for applying the constraint
    """
    frequent_strings = MetricsSelector(metric_name="frequent_items", column_name=column_name)
    constraint_name = f"{column_name} {n}-most common items in set {reference_set}"

    def most_common_in_set(metric):
        frequent_strings = metric.to_summary_dict(SummaryConfig())["frequent_strings"]
        most_common_items = frequent_strings[:n]
        result = all(item.value in reference_set for item in most_common_items)
        return result

    constraint = MetricConstraint(name=constraint_name, condition=most_common_in_set, metric_selector=frequent_strings)
    return constraint



import logging
import os
from tempfile import mkdtemp
from typing import Any, Optional

import mlflow

from whylogs.api.writer import Writer
from whylogs.api.writer.writer import Writable
from whylogs.core.utils import deprecated_alias

logger = logging.getLogger(__name__)


class MlflowWriter(Writer):
    def __init__(self) -> None:
        self._file_dir = "whylogs"
        self._file_name = None
        self._end_run = True

    @deprecated_alias(profile="file")
    def write(
        self,
        file: Writable,
        dest: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        run = mlflow.active_run() or mlflow.start_run()
        self._run_id = run.info.run_id
        dest = dest or self._file_name or file.get_default_path()  # dest has a higher priority than file_name
        output = self._get_temp_directory(dest=dest)
        file.write(path=output)  # type: ignore
        mlflow.log_artifact(output, artifact_path=self._file_dir)

        if self._end_run is True:
            mlflow.end_run()

    @deprecated_alias(profile_dir="file_dir", profile_name="file_name")
    def option(
        self, file_name: Optional[str] = None, file_dir: Optional[str] = None, end_run: Optional[bool] = None
    ) -> None:
        if end_run is not None:
            self._end_run = end_run
        if file_dir:
            self._file_dir = file_dir
        if file_name:
            self._file_name = file_name  # type: ignore

    def _get_temp_directory(self, dest: Optional[str] = None):
        tmp_dir = mkdtemp()
        output_dir = os.path.join(tmp_dir, self._file_dir)
        os.makedirs(output_dir, exist_ok=True)
        if dest:
            output = os.path.join(output_dir, dest)
        elif self._file_name:
            output = os.path.join(output_dir, self._file_name)
        else:
            output = output_dir
        return output
